"""

Owl/tax2026
---

A retirement planner using linear programming optimization.

See companion document for a complete explanation and description
of all variables and parameters.

Module to handle all tax calculations.

Copyright &copy; 2026 - Martin-D. Lacasse

Disclaimers: This code is for educational purposes only and does not constitute financial advice.

"""

import numpy as np
from datetime import date


##############################################################################
# Prepare the data.

taxBracketNames = ["10%", "12/15%", "22/25%", "24/28%", "32/33%", "35%", "37/40%"]

rates_OBBBA = np.array([0.10, 0.12, 0.22, 0.24, 0.32, 0.35, 0.370])
rates_preTCJA = np.array([0.10, 0.15, 0.25, 0.28, 0.33, 0.35, 0.396])

###############################################################################
# Start of section where rates need to be actualized every year.
###############################################################################
# Single [0] and married filing jointly [1].

# These are 2025 current.
taxBrackets_OBBBA = np.array(
    [
        [12400, 50400, 105700, 201775, 256225, 640600, 9999999],
        [24800, 100800, 211400, 403550, 512450, 768700, 9999999],
    ]
)

irmaaBrackets = np.array(
    [
        [0, 106000, 133000, 167000, 200000, 500000],
        [0, 212000, 266000, 334000, 400000, 750000],
    ]
)

# Index [0] stores the standard Medicare part B premium.
# Following values are incremental IRMAA part B monthly fees.
irmaaFees = 12 * np.array([185.00, 74.00, 111.00, 110.90, 111.00, 37.00])

# Make projection for pre-TCJA using 2017 to current year.
# taxBrackets_2017 = np.array(
#    [ [9325, 37950, 91900, 191650, 416700, 418400, 9999999],
#      [18650, 75900, 153100, 233350, 416700, 470700, 9999999],
#    ])
#
# stdDeduction_2017 = [6350, 12700]
#
# For 2025, I used a 30.5% adjustment from 2017, rounded to closest 50.
#
# These are speculated.
taxBrackets_preTCJA = np.array(
    [
        [12150, 49550, 119950, 250200, 544000, 546200, 9999999],  # Single
        [24350, 99100, 199850, 304600, 543950, 614450, 9999999],  # MFJ
    ]
)

# These are 2025 current.
stdDeduction_OBBBA = np.array([16100, 32200])    # Single, MFJ
# These are speculated (adjusted for inflation to 2026). TODO
stdDeduction_preTCJA = np.array([8300, 16600])  # Single, MFJ

# These are current (adjusted for inflation) per individual.
extra65Deduction = np.array([2000, 1600])       # Single, MFJ

# Thresholds setting capital gains brackets 0%, 15%, 20% (adjusted for inflation).
capGainRates = np.array(
    [
        [48350, 533400],
        [96700, 600050],
    ]
)

# Thresholds for net investment income tax (not adjusted for inflation).
niitThreshold = np.array([200000, 250000])
niitRate = 0.038

# Thresholds for 65+ bonus for circumventing tax on social security.
bonusThreshold = np.array([75000, 150000])

###############################################################################
# End of section where rates need to be actualized every year.
###############################################################################


def mediVals(yobs, horizons, gamma_n, Nn, Nq):
    """
    Return tuple (nm, L, C) of year index when Medicare starts and vectors L, and C
    defining end points of constant piecewise linear functions representing IRMAA fees.
    """
    thisyear = date.today().year
    assert Nq == len(irmaaFees), f"Inconsistent value of Nq: {Nq}."
    assert Nq == len(irmaaBrackets[0]), "Inconsistent IRMAA brackets array."
    Ni = len(yobs)
    # What index year will Medicare start? 65 - age.
    nm = 65 - (thisyear - yobs)
    nm = np.min(nm)
    # Has it already started?
    nm = max(0, nm)
    Nmed = Nn - nm

    L = np.zeros((Nmed, Nq-1))
    C = np.zeros((Nmed, Nq))

    # Year starts at offset nm in the plan.
    for nn in range(Nmed):
        imed = 0
        n = nm + nn
        if thisyear + n - yobs[0] >= 65 and n < horizons[0]:
            imed += 1
        if Ni == 2 and thisyear + n - yobs[1] >= 65 and n < horizons[1]:
            imed += 1
        if imed:
            status = 0 if Ni == 1 else 1 if n < horizons[0] and n < horizons[1] else 0
            L[nn] = gamma_n[n] * irmaaBrackets[status][1:]
            C[nn] = imed * gamma_n[n] * irmaaFees
        else:
            raise RuntimeError("mediVals: This should never happen.")

    return nm, L, C


def capitalGainTaxRate(Ni, magi_n, gamma_n, nd, Nn):
    """
    Return an array of decimal rates for capital gains.
    Parameter nd is the index year of first passing of a spouse, if applicable,
    nd == Nn for single individuals.
    """
    status = Ni - 1
    cgRate_n = np.zeros(Nn)

    for n in range(Nn):
        if status and n == nd:
            status -= 1

        if magi_n[n] > gamma_n[n] * capGainRates[status][1]:
            cgRate_n[n] = 0.20
        elif magi_n[n] > gamma_n[n] * capGainRates[status][0]:
            cgRate_n[n] = 0.15

    return cgRate_n


def mediCosts(yobs, horizons, magi, prevmagi, gamma_n, Nn):
    """
    Compute Medicare costs directly.
    """
    thisyear = date.today().year
    Ni = len(yobs)
    costs = np.zeros(Nn)
    for n in range(Nn):
        status = 0 if Ni == 1 else 1 if n < horizons[0] and n < horizons[1] else 0
        for i in range(Ni):
            if thisyear + n - yobs[i] >= 65 and n < horizons[i]:
                # Start with the (inflation-adjusted) basic Medicare part B premium.
                costs[n] += gamma_n[n] * irmaaFees[0]
                if n < 2:
                    mymagi = prevmagi[n]
                else:
                    mymagi = magi[n - 2]
                for q in range(1, 6):
                    if mymagi > gamma_n[n] * irmaaBrackets[status][q]:
                        costs[n] += gamma_n[n] * irmaaFees[q]

    return costs


def taxParams(yobs, i_d, n_d, N_n, gamma_n, MAGI_n, yOBBBA=2099):
    """
    Input is year of birth, index of shortest-lived individual,
    lifespan of shortest-lived individual, total number of years
    in the plan, and the year that preTCJA rates might come back.

    It returns 3 time series:
    1) Standard deductions at year n (sigma_n).
    2) Tax rate in year n (theta_tn)
    3) Delta from top to bottom of tax brackets (Delta_tn)
    This is pure speculation on future values.
    Returned values are not indexed for inflation.
    """
    # Compute the deltas in-place between brackets, starting from the end.
    deltaBrackets_OBBBA = np.array(taxBrackets_OBBBA)
    deltaBrackets_preTCJA = np.array(taxBrackets_preTCJA)
    for t in range(6, 0, -1):
        for i in range(2):
            deltaBrackets_OBBBA[i, t] -= deltaBrackets_OBBBA[i, t - 1]
            deltaBrackets_preTCJA[i, t] -= deltaBrackets_preTCJA[i, t - 1]

    # Prepare the 3 arrays to return - use transpose for easy slicing.
    sigmaBar = np.zeros((N_n))
    Delta = np.zeros((N_n, 7))
    theta = np.zeros((N_n, 7))

    filingStatus = len(yobs) - 1
    souls = list(range(len(yobs)))
    thisyear = date.today().year

    for n in range(N_n):
        # First check if shortest-lived individual is still with us.
        if n == n_d:
            souls.remove(i_d)
            filingStatus -= 1

        if thisyear + n < yOBBBA:
            sigmaBar[n] = stdDeduction_OBBBA[filingStatus] * gamma_n[n]
            Delta[n, :] = deltaBrackets_OBBBA[filingStatus, :]
        else:
            sigmaBar[n] = stdDeduction_preTCJA[filingStatus] * gamma_n[n]
            Delta[n, :] = deltaBrackets_preTCJA[filingStatus, :]

        # Add 65+ additional exemption(s) and "bonus" phasing out.
        for i in souls:
            if thisyear + n - yobs[i] >= 65:
                sigmaBar[n] += extra65Deduction[filingStatus] * gamma_n[n]
                if thisyear + n <= 2028:
                    sigmaBar[n] += 6000 * max(0, 1 - 0.06*max(0, MAGI_n[n] - bonusThreshold[filingStatus]))

        # Fill in future tax rates for year n.
        if thisyear + n < yOBBBA:
            theta[n, :] = rates_OBBBA[:]
        else:
            theta[n, :] = rates_preTCJA[:]

    Delta = Delta.transpose()
    theta = theta.transpose()

    # Return series unadjusted for inflation, except for sigmaBar, in STD order.
    return sigmaBar, theta, Delta


def taxBrackets(N_i, n_d, N_n, yOBBBA=2099):
    """
    Return dictionary containing future tax brackets
    unadjusted for inflation for plotting.
    """
    if not (0 < N_i <= 2):
        raise ValueError(f"Cannot process {N_i} individuals.")

    n_d = min(n_d, N_n)
    status = N_i - 1

    # Number of years left in OBBBA from this year.
    thisyear = date.today().year
    if yOBBBA < thisyear:
        raise ValueError(f"Expiration year {yOBBBA} cannot be in the past.")

    ytc = yOBBBA - thisyear

    data = {}
    for t in range(len(taxBracketNames) - 1):
        array = np.zeros(N_n)
        for n in range(N_n):
            stat = status if n < n_d else 0
            array[n] = taxBrackets_OBBBA[stat][t] if n < ytc else taxBrackets_preTCJA[stat][t]

        data[taxBracketNames[t]] = array

    return data


def rho_in(yobs, N_n):
    """
    Return Required Minimum Distribution fractions for each individual.
    This implementation does not support spouses with more than
    10-year difference.
    It starts at age 73 until it goes to 75 in 2033.
    """
    # Notice that table starts at age 72.
    rmdTable = [
        27.4,
        26.5,
        25.5,
        24.6,
        23.7,
        22.9,
        22.0,
        21.1,
        20.2,
        19.4,
        18.5,
        17.7,
        16.8,
        16.0,
        15.2,
        14.4,
        13.7,
        12.9,
        12.2,
        11.5,
        10.8,
        10.1,
        9.5,
        8.9,
        8.4,
        7.8,
        7.3,
        6.8,
        6.4,
        6.0,
        5.6,
        5.2,
        4.9,
        4.6,
    ]

    N_i = len(yobs)
    if N_i == 2 and abs(yobs[0] - yobs[1]) > 10:
        raise RuntimeError("RMD: Unsupported age difference of more than 10 years.")

    rho = np.zeros((N_i, N_n))
    thisyear = date.today().year
    for i in range(N_i):
        agenow = thisyear - yobs[i]
        # Account for increase of RMD age between 2023 and 2032.
        yrmd = 70 if yobs[i] < 1949 else 72 if 1949 <= yobs[i] <= 1950 else 73 if 1951 <= yobs[i] <= 1959 else 75
        for n in range(N_n):
            yage = agenow + n

            if yage < yrmd:
                pass  # rho[i][n] = 0
            else:
                rho[i][n] = 1.0 / rmdTable[yage - 72]

    return rho



"""
Plotting backends for Owl.

Copyright &copy; 2025 - Martin-D. Lacasse

Disclaimers: This code is for educatonal purposes only and does not constitute financial advice.

"""

from .factory import PlotFactory

__all__ = ['PlotFactory']



import streamlit as st

import sskeys as kz
import owlbridge as owb


ret = kz.titleBar(":material/stacked_line_chart: Graphs")

if ret is None or kz.caseHasNoPlan():
    st.info("Case(s) must be first created before running this page.")
else:
    if kz.caseIsRunReady():
        owb.runPlan()
    elif kz.caseHasNotRun():
        st.info("Case definition is not yet complete. Please visit all pages in *Case Setup*.")

    st.write("Optimize a single scenario based on the parameters selected in the **Case Setup** section.")
    col1, col2, col3 = st.columns(3, gap="large", vertical_alignment="bottom")
    with col1:
        choices = ["nominal", "today"]
        kz.initCaseKey("plots", choices[0])
        helpmsg = "Plot can be in today's dollars or in nominal value."
        ret = kz.getRadio("Dollar amounts in plots", choices, "plots", help=helpmsg,
                          callback=owb.setDefaultPlots)

    with col3:
        helpmsg = "Click on button if graphs are not all showing."
        st.button(
            "Re-run single case",
            help=helpmsg,
            on_click=owb.runPlan,
            disabled=kz.caseIsNotRunReady(),
        )

    st.divider()
    if kz.isCaseUnsolved():
        st.info("Case status is currently '%s'." % kz.getCaseKey("caseStatus"))
    else:
        owb.plotSingleResults()



"""
Base classes for plot backends.

Copyright &copy; 2025 - Martin-D. Lacasse

Disclaimers: This code is for educatonal purposes only and does not constitute financial advice.

"""

from abc import ABC, abstractmethod


class PlotBackend(ABC):
    """Abstract base class for plot backends."""

    @abstractmethod
    def jupyter_renderer(self, fig):
        """Render plot for Jupyter."""
        pass

    @abstractmethod
    def plot_histogram_results(self, objective, df, N, year_n, n_d=None, N_i=1, phi_j=None):
        """Show a histogram of values from historical data or Monte Carlo simulations."""
        pass

    @abstractmethod
    def plot_rates_correlations(self, name, tau_kn, N_n, rate_method, rate_frm=None, rate_to=None,
                                tag="", share_range=False):
        """Plot correlations between various rates."""
        pass

    @abstractmethod
    def plot_rates(self, name, tau_kn, year_n, N_k, rate_method,
                   rate_frm=None, rate_to=None, tag=""):
        """Plot rate values used over the time horizon."""
        pass

    @abstractmethod
    def plot_rates_distributions(self, frm, to, SP500, BondsBaa, TNotes, Inflation, FROM):
        """Plot histograms of the rates distributions."""
        pass

    @abstractmethod
    def plot_gross_income(self, year_n, G_n, gamma_n, value, title, tax_brackets):
        """Plot gross income over time."""
        pass

    @abstractmethod
    def plot_profile(self, year_n, xi_n, title, inames):
        """Plot profile over time."""
        pass

    @abstractmethod
    def plot_net_spending(self, year_n, g_n, xi_n, xiBar_n, gamma_n, value, title, inames):
        """Plot net spending over time."""
        pass

    @abstractmethod
    def plot_asset_composition(self, year_n, inames, b_ijkn, gamma_n, value, name, tag):
        """Plot asset composition over time."""
        pass

    @abstractmethod
    def plot_allocations(self, year_n, inames, alpha_ijkn, ARCoord, title):
        """Plot allocations over time."""
        pass

    @abstractmethod
    def plot_accounts(self, year_n, savings_in, gamma_n, value, title, inames):
        """Plot accounts over time."""
        pass

    @abstractmethod
    def plot_sources(self, year_n, sources_in, gamma_n, value, title, inames):
        """Plot sources over time."""
        pass

    @abstractmethod
    def plot_taxes(self, year_n, T_n, M_n, gamma_n, value, title, inames):
        """Plot taxes over time."""
        pass



from io import StringIO
import streamlit as st

import sskeys as kz
import owlbridge as owb
import tomlexamples as tomlex


ret = kz.titleBar(":material/person_add: Create Case", allCases=True)

if ret == kz.newCase:
    st.info("#### Starting a new case from scratch.\n\n" "A name for the scenario must first be provided.")
    st.text_input(
        "Case name",
        value="",
        key="_newcase",
        on_change=kz.createNewCase,
        args=["newcase"],
        placeholder="Enter a case name...",
    )
elif ret == kz.loadCaseFile:
    st.info(
        "#### Starting a case from a *case* parameter file.\n\n"
        "Upload your own case or select one from multiple examples."
        " Alternatively, you can select `New Case...` in the top selector box to start a case from scratch.\n\n"
        "Look at the :material/help: [Documentation](Documentation) for more details."
    )
    col1, col2 = st.columns(2, gap="large")
    with col1:
        st.write("#### :orange[Upload Your Own Case File]")
        file = st.file_uploader("Upload *case* parameter file...", key="_confile", type=["toml"])
        if file is not None:
            mystringio = StringIO(file.read().decode("utf-8"))
            if kz.createCaseFromFile(mystringio):
                st.rerun()

    with col2:
        st.write("#### :orange[Load a Case Example]")
        case = st.selectbox("Examples available from GitHub", tomlex.cases, index=None, placeholder="Select a case")
        if case:
            mystringio = tomlex.loadCaseExample(case)
            if kz.createCaseFromFile(mystringio):
                kz.initCaseKey("tomlexcase", case)
                st.rerun()
else:
    st.write("#### :orange[Description and Life Parameters]")
    helpmsg = "Case name can be changed by editing it directly."
    col1, col2 = st.columns(2, gap="large")
    with col1:
        name = st.text_input(
            "Case name",
            value=kz.currentCaseName(),
            on_change=kz.renameCase,
            args=["caseNewName"],
            key="caseNewName",
            help=helpmsg,
        )

    diz1 = kz.getCaseKey("plan") is not None
    diz2 = diz1
    # diz2 = (diz1 or len(kz.allCaseNames()) > 3)
    with col2:
        statusChoices = ["single", "married"]
        kz.initCaseKey("status", statusChoices[0])
        st.radio(
            "Marital status",
            statusChoices,
            disabled=diz2,
            index=statusChoices.index(kz.getCaseKey("status")),
            key=kz.genCaseKey("status"),
            on_change=kz.setpull,
            args=["status"],
            horizontal=True,
        )

    kz.initCaseKey("description", "")
    helpmsg = "Provide a short distinguishing description for the case."
    description = kz.getLongText("Brief description", "description", help=helpmsg,
                                 placeholder="Enter a brief description...")

    col1, col2 = st.columns(2, gap="large", vertical_alignment="top")
    with col1:
        kz.initCaseKey("iname0", "")
        if kz.getCaseKey("iname0") == "":
            st.info("First name must be provided.")

        iname0 = kz.getText("Your first name", "iname0",
                            disabled=diz2, placeholder="Enter name...")

        if iname0:
            kz.initCaseKey("yob0", 1965)
            ret = kz.getIntNum(f"{iname0}'s birth year", "yob0", disabled=diz2)

            kz.initCaseKey("life0", 80)
            ret = kz.getIntNum(f"{iname0}'s expected longevity", "life0", disabled=diz1)

    with col2:
        if kz.getCaseKey("status") == "married":
            kz.initCaseKey("iname1", "")
            if kz.getCaseKey("iname1") == "":
                st.info("First name must be provided.")

            iname1 = kz.getText("Your spouse's first name", "iname1", disabled=diz2, placeholder="Enter a name...")

            if iname1:
                kz.initCaseKey("yob1", 1965)
                ret = kz.getIntNum(f"{iname1}'s birth year", "yob1", disabled=diz2)

                kz.initCaseKey("life1", 80)
                ret = kz.getIntNum(f"{iname1}'s expected longevity", "life1", disabled=diz1)

    st.divider()
    cantcreate = kz.isIncomplete() or diz1
    if not cantcreate and kz.getCaseKey("plan") is None:
        st.info("Plan needs to be created once desired changes are completed.")

    cantmodify = kz.currentCaseName() == kz.newCase or kz.currentCaseName() == kz.loadCaseFile
    cantcopy = cantmodify or kz.caseHasNoPlan()
    col1, col2, col3 = st.columns(3, gap="small", vertical_alignment="top")
    with col1:
        st.button("Create case :material/add:", on_click=owb.createPlan, disabled=cantcreate)
    with col2:
        st.button("Duplicate case :material/content_copy:", on_click=kz.duplicateCase, disabled=cantcopy)
    with col3:
        st.button("Delete case :material/delete:", on_click=kz.deleteCurrentCase, disabled=cantmodify)



import pandas as pd
import ui.sskeys as sskeys


def test_compareSummaries(monkeypatch):
    # Setup mock session state and cases
    df1 = pd.DataFrame({'A': ['$100'], 'B': ['$200']})
    df2 = pd.DataFrame({'A': ['$110'], 'B': ['$190']})
    cases = {
        "case1": {"summaryDf": df1},
        "case2": {"summaryDf": df2},
    }
    monkeypatch.setattr(sskeys, "getCaseKey", lambda key: df1 if key == "summaryDf" else None)
    monkeypatch.setattr(sskeys, "onlyCaseNames", lambda: ["case2"])
    monkeypatch.setattr(sskeys, "currentCaseName", lambda: "case1")
    sskeys.ss = type("SS", (), {"cases": cases})()
    result = sskeys.compareSummaries()
    assert result is not None



from io import StringIO

from owlplanner import mylogging as log


def test_logger1():
    strio = StringIO()
    mylog = log.Logger(True, [strio])
    msg1 = 'Hello'
    mylog.vprint(msg1)
    msg2 = strio.getvalue().splitlines()
    assert msg1 == msg2[-1]


def test_logger2():
    strio = StringIO()
    mylog = log.Logger(False, [strio])
    msg1 = 'Hello'
    mylog.vprint(msg1)
    msg2 = strio.getvalue()
    assert '' == msg2




import pytest
import os
from datetime import date
from io import StringIO

import owlplanner as owl


# solver = 'MOSEK'
solver = 'HiGHS'
thisyear = date.today().year


def createJackAndJillPlan(name):
    inames = ['Jack', 'Jill']
    yobs = [1963, 1966]
    expectancy = [thisyear - 1963 + 20, thisyear - 1966 + 20]
    p = owl.Plan(inames, yobs, expectancy, name)
    p.setSpendingProfile('flat', 60)

    p.setAccountBalances(taxable=[90, 60], taxDeferred=[600, 150],
                         taxFree=[50 + 20, 40], startDate="1-1")
    p.readContributions('./examples/jack+jill.xlsx')
    p.setInterpolationMethod('s-curve')
    p.setAllocationRatios('individual',
                          generic=[[[60, 40, 0, 0], [70, 30, 0, 0]],
                                   [[50, 50, 0, 0], [70, 30, 0, 0]]])
    p.setPension([0, 10], [65, 65])
    p.setSocialSecurity([28, 25], [70, 70])

    return p


def test_case1():
    p = createJackAndJillPlan('case1')
    p.setRates('historical', 1969)
    p.solve('maxSpending', options={'maxRothConversion': 100, 'bequest': 500})
    assert p.caseStatus == "solved"
    assert p.basis == pytest.approx(83629.1, abs=0.5)
    assert p.bequest == pytest.approx(500000, abs=0.5)


def test_case2():
    p = createJackAndJillPlan('case2')
    p.setRates('historical', 1969)
    p.solve('maxBequest', options={'maxRothConversion': 100, 'netSpending': 80})
    assert p.caseStatus == "solved"
    assert p.basis == pytest.approx(80000, abs=0.5)
    assert p.bequest == pytest.approx(675352.7, abs=0.5)


def test_config1():
    name = 'testconfig'
    p = createJackAndJillPlan(name)
    p.setRates('historical', 1969)
    p.solve('maxBequest', options={'maxRothConversion': 100, 'netSpending': 80})
    assert p.caseStatus == "solved"
    assert p.basis == pytest.approx(80000, abs=0.5)
    assert p.bequest == pytest.approx(675352.7, abs=0.5)
    p.saveConfig()
    base_filename = 'case_' + name
    full_filename = 'case_' + name + '.toml'
    assert os.path.isfile(full_filename)
    p2 = owl.readConfig(base_filename)
    p2.solve('maxBequest', options={'maxRothConversion': 100, 'netSpending': 80})
    assert p2.caseStatus == "solved"
    assert p2.basis == pytest.approx(80000, abs=0.5)
    assert p2.bequest == pytest.approx(675352.7, abs=0.5)
    p3 = owl.readConfig(full_filename)
    p3.solve('maxBequest', options={'maxRothConversion': 100, 'netSpending': 80})
    assert p3.caseStatus == "solved"
    assert p3.basis == pytest.approx(80000, abs=0.5)
    assert p3.bequest == pytest.approx(675352.7, abs=0.5)
    os.remove(full_filename)


def test_config2():
    name = 'testconfig'
    p = createJackAndJillPlan(name)
    p.setRates('historical', 1969)
    p.solve('maxBequest', options={'maxRothConversion': 100, 'netSpending': 80})
    assert p.caseStatus == "solved"
    assert p.basis == pytest.approx(80000, abs=0.5)
    assert p.bequest == pytest.approx(675352.7, abs=0.5)
    iostring = StringIO()
    p.saveConfig(iostring)
    # print('iostream:', iostream.getvalue())
    p2 = owl.readConfig(iostring)
    p2.solve('maxBequest', options={'maxRothConversion': 100, 'netSpending': 80})
    assert p2.caseStatus == "solved"
    assert p2.basis == pytest.approx(80000, abs=0.5)
    assert p2.bequest == pytest.approx(675352.7, abs=0.5)


def test_clone1():
    name = 'testclone1.1'
    p = createJackAndJillPlan(name)
    p.setRates('historical', 1969)
    p.solve('maxBequest', options={'maxRothConversion': 100, 'netSpending': 80})
    assert p.caseStatus == "solved"
    assert p.basis == pytest.approx(80000, abs=0.5)
    assert p.bequest == pytest.approx(675352.7, abs=0.5)
    name2 = 'testclone1.2'
    p2 = owl.clone(p, name2)
    p2.solve('maxBequest', options={'maxRothConversion': 100, 'netSpending': 80})
    assert p2.caseStatus == "solved"
    assert p2.basis == pytest.approx(80000, abs=0.5)
    assert p2.bequest == pytest.approx(675352.7, abs=0.5)


def test_clone2():
    name = 'testclone2.1'
    p = createJackAndJillPlan(name)
    p.setRates('historical', 1969)
    p.solve('maxSpending', options={'maxRothConversion': 100, 'bequest': 10})
    assert p.caseStatus == "solved"
    assert p.basis == pytest.approx(93760.0, abs=0.5)
    assert p.bequest == pytest.approx(10000, abs=0.5)
    name2 = 'testclone2.2'
    p2 = owl.clone(p, name2)
    p2.solve('maxSpending', options={'maxRothConversion': 100, 'bequest': 10})
    assert p2.caseStatus == "solved"
    assert p2.basis == pytest.approx(93760.0, abs=0.5)
    assert p2.bequest == pytest.approx(10000, abs=0.5)



from datetime import date
import streamlit as st

import sskeys as kz

ret = kz.titleBar(":material/savings: Current Assets")

if ret is None or kz.caseHasNoPlan():
    st.info("Case(s) must be first created before running this page.")
else:
    st.write("#### :orange[Savings Account Balances]")
    accounts = {"txbl": "taxable", "txDef": "tax-deferred", "txFree": "tax-free"}
    hdetails = {"txbl": "Brokerage and savings accounts excluding emergency fund. ",
                "txDef": "IRA, 401k, 403b and the like. ",
                "txFree": "Roth IRA, Roth 401k, Roth 403b and the like. "}
    col1, col2, col3 = st.columns(3, gap="large", vertical_alignment="top")
    with col1:
        iname = kz.getCaseKey("iname0")
        for key in accounts:
            nkey = key + str(0)
            kz.initCaseKey(nkey, 0)
            ret = kz.getNum(f"{iname}'s {accounts[key]} account ($k)", nkey,
                            help=hdetails[key]+kz.help1000)

        today = date.today()
        thisyear = today.year
        kz.initCaseKey("startDate", today)
        helpmsg = "Date at which savings balances are known. Values will be back projected to Jan 1st."
        ret = st.date_input("Account balance date", min_value=date(thisyear, 1, 1),
                            max_value=date(thisyear, 12, 31), value=kz.getCaseKey("startDate"),
                            key=kz.genCaseKey("startDate"), args=["startDate"], on_change=kz.setpull, help=helpmsg)

    with col2:
        if kz.getCaseKey("status") == "married":
            iname1 = kz.getCaseKey("iname1")
            for key in accounts:
                nkey = key + str(1)
                kz.initCaseKey(nkey, 0)
                ret = kz.getNum(f"{iname1}'s {accounts[key]} account ($k)", nkey,
                                help=hdetails[key]+kz.help1000)

    if kz.getCaseKey("status") == "married":
        st.divider()
        st.write("#### :orange[Survivor's Spousal Beneficiary Fractions]")
        col1, col2, col3 = st.columns(3, gap="large", vertical_alignment="top")
        with col1:
            nkey = "benf" + str(0)
            kz.initCaseKey(nkey, 1)
            helpmsg = "Fraction of account left to surviving spouse."
            ret = kz.getNum(accounts["txbl"].capitalize(), nkey, format="%.2f", max_value=1.0,
                            step=0.05, help=helpmsg)

        with col2:
            nkey = "benf" + str(1)
            kz.initCaseKey(nkey, 1)
            ret = kz.getNum(accounts["txDef"].capitalize(), nkey, format="%.2f", max_value=1.0,
                            step=0.05, help=helpmsg)

        with col3:
            nkey = "benf" + str(2)
            kz.initCaseKey(nkey, 1)
            ret = kz.getNum(accounts["txFree"].capitalize(), nkey, format="%.2f", max_value=1.0,
                            step=0.05, help=helpmsg)

        st.write("#####")
        st.write("#### :orange[Surplus Deposit Fraction]")
        col1, col2, col3 = st.columns(3, gap="large", vertical_alignment="top")
        with col1:
            kz.initCaseKey("surplusFraction", 0.5)
            helpmsg = ("When beneficiary fractions not all 1, "
                       "set surplus deposits to all go to account of first spouse to pass.")
            ret = kz.getNum(
                f"Fraction deposited in {iname1}'s taxable account",
                "surplusFraction",
                format="%.2f",
                help=helpmsg,
                max_value=1.0,
                step=0.05,
            )



import streamlit as st
from datetime import date

import sskeys as kz
import owlbridge as owb


profileChoices = ["flat", "smile"]
kz.initCaseKey("spendingProfile", profileChoices[1])
kz.initCaseKey("survivor", 60)
kz.initCaseKey("smileDip", 15)
kz.initCaseKey("smileIncrease", 12)
kz.initCaseKey("smileDelay", 0)
mediChoices = ["None", "loop", "optimize"]
kz.initCaseKey("withMedicare", mediChoices[1])


def initProfile():
    owb.setProfile(None)


ret = kz.titleBar(":material/tune: Optimization Parameters")

if ret is None or kz.caseHasNoPlan():
    st.info("Case(s) must be first created before running this page.")
else:
    kz.runOncePerCase(initProfile)

    st.write("#### :orange[Objective]")
    col1, col2 = st.columns(2, gap="large", vertical_alignment="top")
    with col1:
        choices = ["Net spending", "Bequest"]
        helpmsg = "Value is in today's \\$k."
        kz.initCaseKey("objective", choices[0])
        ret = kz.getRadio("Maximize", choices, "objective")

    with col2:
        if kz.getCaseKey("objective") == "Net spending":
            kz.initCaseKey("bequest", 0)
            ret = kz.getNum("Desired bequest (\\$k)", "bequest", help=helpmsg)

        else:
            kz.initCaseKey("netSpending", 0)
            ret = kz.getNum("Desired annual net spending (\\$k)", "netSpending", help=helpmsg)

    st.divider()
    st.write("#### :orange[Roth Conversions]")
    col1, col2 = st.columns(2, gap="large", vertical_alignment="top")
    with col1:
        iname0 = kz.getCaseKey("iname0")
        helpmsg = "Value is in nominal \\$k."
        kz.initCaseKey("readRothX", False)
        fromFile = kz.getCaseKey("readRothX")
        kz.initCaseKey("maxRothConversion", 50)
        ret = kz.getNum("Maximum annual Roth conversion (\\$k)", "maxRothConversion",
                        disabled=fromFile, help=helpmsg)
        ret = kz.getToggle("Convert as in Wages and Contributions tables", "readRothX")
        # kz.initCaseKey("oppCostX", 0.)
        # helpmsg = "Estimated opportunity cost for paying estimated tax on Roth conversions."
        # ret = kz.getNum("Opportunity cost for conversion (%)", "oppCostX", step=0.01, format="%.2f",
        #                min_value=0., max_value=5., help=helpmsg)

    with col2:
        helpmsg = "Do not perform Roth conversions before that year."
        thisyear = date.today().year
        kz.initCaseKey("startRothConversions", thisyear)
        ret = kz.getIntNum("Year to start considering Roth conversions", "startRothConversions",
                           min_value=thisyear, disabled=fromFile, help=helpmsg)
        if kz.getCaseKey("status") == "married":
            iname1 = kz.getCaseKey("iname1")
            choices = ["None", iname0, iname1]
            kz.initCaseKey("noRothConversions", choices[0])
            helpmsg = "`None` means no exclusion. To exclude both spouses, set `Maximum Roth conversion` to 0."
            ret = kz.getRadio("Exclude Roth conversions for...", choices, "noRothConversions", help=helpmsg)

    st.divider()
    st.write("#### :orange[Calculations]")
    kz.initCaseKey("withSCLoop", True)
    kz.initCaseKey("xorConstraints", True)
    col1, col2 = st.columns(2, gap="large", vertical_alignment="top")
    with col1:
        helpmsg = ("Option to use a self-consistent loop to adjust additional values such as the net"
                   " investment income tax (NIIT), and capital gain tax rates."
                   "  If selected below, this loop will also compute Medicare and IRMAA.")
        ret = kz.getToggle("Self-consistent loop calculations", "withSCLoop", help=helpmsg)
    with col2:
        helpmsg = ("Enable mutually exclusive constraints between surplus deposits,"
                   " Roth conversions, and withdrawals from taxable and/or tax-free accounts.")
        ret = kz.getToggle("XOR constraints on deposits, conversions, and withdrawals",
                           "xorConstraints", help=helpmsg)

    st.divider()
    st.write("#### :orange[Medicare]")
    col1, col2 = st.columns(2, gap="large", vertical_alignment="top")
    with col1:
        helpmsg = ("How to compute Medicare and IRMAA premiums:"
                   " ignore, use self-consistent loop, or use additional variables in optimization.")
        ret = kz.getRadio("Medicare and IRMAA calculations", mediChoices, "withMedicare", help=helpmsg)
        if ret == "optimize":
            st.markdown(":material/warning: Medicare optimization can sometimes have slow convergence -"
                        " time for :coffee: ?")
        elif ret == "loop" and not kz.getCaseKey("withSCLoop"):
            st.markdown(":material/warning: Medicare set to 'loop' while self-consistent loop is off.")
    with col2:
        medi = kz.getCaseKey("withMedicare")
        if medi == "optimize" or (medi == "loop" and kz.getCaseKey("withSCLoop")):
            helpmsg = "MAGI in nominal $k for current and previous years."
            years = owb.backYearsMAGI()
            for ii in range(2):
                kz.initCaseKey("MAGI" + str(ii), 0)
                if years[ii] > 0:
                    ret = kz.getNum(f"MAGI for year {years[ii]} ($k)", "MAGI" + str(ii), help=helpmsg)

    st.divider()
    st.write("#### :orange[Solver]")
    choices = ["HiGHS", "PuLP/CBC", "PuLP/HiGHS"]
    if owb.hasMOSEK():
        choices += ["MOSEK"]
    kz.initCaseKey("solver", choices[0])
    helpmsg = "Select different solvers for comparison purposes. Use HiGHS for best performance."
    ret = kz.getRadio("Linear programming solver", choices, "solver", help=helpmsg)

    st.divider()
    st.write("#### :orange[Spending Profile]")
    col1, col2, col3 = st.columns(3, gap="medium", vertical_alignment="top")
    with col1:
        ret = kz.getRadio("Type of profile", profileChoices, "spendingProfile", callback=owb.setProfile)
        if kz.getCaseKey("spendingProfile") == "smile":
            helpmsg = "Time in year before spending starts decreasing."
            ret = kz.getIntNum("Smile delay (in years from now)", "smileDelay", max_value=30,
                               help=helpmsg, callback=owb.setProfile)
    with col2:
        kz.initCaseKey("spendingSlack", 0)
        helpmsg = "Percentage allowed to deviate from spending profile."
        ret = kz.getIntNum("Profile slack (%)", "spendingSlack", max_value=50, help=helpmsg)
        if kz.getCaseKey("spendingProfile") == "smile":
            helpmsg = "Percentage to decrease for the slow-go years."
            ret = kz.getIntNum("Smile dip (%)", "smileDip", max_value=100, help=helpmsg,
                               callback=owb.setProfile)
    with col3:
        if kz.getCaseKey("status") == "married":
            helpmsg = "Percentage of spending required for the surviving spouse."
            ret = kz.getIntNum("Survivor's spending (%)", "survivor", max_value=100,
                               help=helpmsg, callback=owb.setProfile)
        if kz.getCaseKey("spendingProfile") == "smile":
            helpmsg = "Percentage to increase (or decrease) over time period."
            ret = kz.getIntNum("Smile increase (%)", "smileIncrease",
                               min_value=-100, max_value=100, help=helpmsg, callback=owb.setProfile)

    st.divider()
    col1, col2 = st.columns([0.6, 0.4], gap="small")
    with col1:
        owb.showProfile(col1)



import streamlit as st

import sskeys as kz


ret = kz.titleBar(":material/error: Logs")

if ret is None or kz.caseHasNoPlan():
    st.info("Case(s) must be first created before running this page.")
else:
    strio = kz.getCaseKey("logs")
    if strio is not None:
        logmsg = strio.getvalue()
        st.code(logmsg, language=None)



import streamlit as st

import sskeys as kz
import owlbridge as owb


ret = kz.titleBar(":material/history: Historical Range")

if ret is None or kz.caseHasNoPlan():
    st.info("Case(s) must be first created before running this page.")
else:
    kz.initCaseKey("hyfrm", owb.FROM)
    kz.initCaseKey("hyto", owb.TO)
    kz.initCaseKey("histoPlot", None)
    kz.initCaseKey("histoSummary", None)

    st.write(
        "Generate a histogram of results obtained from backtesting "
        "current scenario with historical data over selected year range."
    )
    col1, col2, col3, col4 = st.columns(4, gap="large", vertical_alignment="bottom")
    with col1:
        st.number_input(
            "Starting year",
            min_value=owb.FROM,
            max_value=kz.getCaseKey("hyto"),
            value=kz.getCaseKey("hyfrm"),
            on_change=kz.storepull,
            args=["hyfrm"],
            key=kz.genCaseKey("hyfrm"),
        )

    with col2:
        st.number_input(
            "Ending year",
            max_value=owb.TO,
            min_value=kz.getCaseKey("hyfrm"),
            value=kz.getCaseKey("hyto"),
            on_change=kz.storepull,
            args=["hyto"],
            key=kz.genCaseKey("hyto"),
        )

    # st.divider()
    # col1, col2 = st.columns(2, gap="small", vertical_alignment="top")
    with col4:
        st.button("Run historical range", on_click=owb.runHistorical, disabled=kz.caseIsNotRunReady())

    st.divider()
    fig = kz.getCaseKey("histoPlot")
    if fig:
        col1, col2 = st.columns(2, gap="medium")
        owb.renderPlot(fig, col1)
        col2.code(kz.getCaseKey("histoSummary"), language=None)



import streamlit as st

import sskeys as kz
import owlbridge as owb

ret = kz.titleBar(":material/data_table: Worksheets")

if ret is None or kz.caseHasNoPlan():
    st.info("Case(s) must be first created before running this page.")
else:
    if kz.caseIsRunReady():
        owb.runPlan()
    elif kz.caseHasNotRun():
        st.info("Case definition is not yet complete. Please visit all pages in *Case Setup*.")

    if kz.isCaseUnsolved():
        st.info("Case status is currently '%s'." % kz.getCaseKey("caseStatus"))
    else:
        owb.showWorkbook()



import streamlit as st

import sskeys as kz

col1, col2, col3 = st.columns([0.69, 0.02, 0.29], gap="large")
with col1:
    st.write("# :material/help: Documentation")
    kz.divider("orange")
    st.write("## :orange[Owl Retirement Planner]\n-------")
with col3:
    st.image("http://github.com/mdlacasse/Owl/blob/main/docs/images/owl.png?raw=true")
    st.caption("Retirement planner with great wisdom")

col1, col2 = st.columns([0.80, 0.20], gap="large")
with col1:
    st.markdown("""
#### A retirement financial exploration tool based on linear programming

The goal of Owl is to provide a free and open-source ecosystem that has cutting-edge
optimization capabilities, allowing for the new generation of computer-literate retirees
to experiment with their own financial future while providing a codebase where they can learn and contribute.
Strictly speaking, Owl is not a planning tool, but more an environment for exploring *what if* scenarios.
It provides different realizations of a financial strategy through the rigorous
mathematical optimization of relevant decision variables. Using a linear programming approach,
two different objectives can currently be optimized: either
maximize the net spending amount under the constraint of a desired bequest,
or maximize an after-tax bequest under the constraint of a desired net spending amount.
In each case, Roth conversions are optimized to reduce the tax burden,
while federal income tax and Medicare premiums (including IRMAA) are calculated.
A full description of the package can be found on the GitHub
open [repository](https://github.com/mdlacasse/Owl), and the mathematical
formulation of the optimization problem can be found
[here](https://github.com/mdlacasse/Owl/blob/main/docs/owl.pdf).

-------------------------------------------------
### :orange[Table of Contents]
[Getting Started with User Interface](#getting-started-with-user-interface)

[Case Setup](#case-setup)
- [:material/person_add: Create Case](#person-add-create-case)
- [:material/work_history: Wages and Contributions](#work-history-wages-and-contributions)
- [:material/currency_exchange: Fixed Income](#currency-exchange-fixed-income)
- [:material/savings: Current Assets](#saving-current-assets)
- [:material/percent: Asset Allocation](#percent-asset-allocation)
- [:material/monitoring: Rates Selection](#monitoring-rates-selection)
- [:material/tune: Optimization Parameters](#tune-optimization-parameters)

[Single Scenario](#single-scenario)
- [:material/stacked_line_chart: Graphs](#stacked-line-chart-graphs)
- [:material/data_table: Worksheets](#data-table-worksheets)
- [:material/description: Output Files](#description-output-files)

[Multiple Scenarios](#multiple-scenarios)
- [:material/history: Historical Range](#history-historical-range)
- [:material/finance: Monte Carlo](#finance-monte-carlo)

[Resources](#resources)
- [:material/rocket_launch: Quick Start](#rocket-launch-quick-start)
- [:material/help: Documentation](#help-documentation)
- [:material/settings: Settings](#settings-settings)
- [:material/error: Logs](#error-logs)
- [:material/info: About Owl](#info-about-owl)

[Tips](#tips)
- [:material/lightbulb_2: Advice on Optimization and Roth Conversions]\
(#lightbulb-2-advice-on-optimization-and-roth-conversions)
- [:material/rule_settings: Typical Workflow](#rule-settings-typical-workflow)
- [:material/mindfulness: Scope of Use](#mindfulness-scope-of-use)

--------------------------------------------------------------------------------------
### :orange[Getting Started with User Interface]
Functions of each page are described below in the same order as they appear in the menu bar:
Typically, pages would be accessed in order, starting from the top.

A `Case selector` box located at the top of the page allows to
navigate between the different scenarios created.
When on the [Create Case](#person-add-create-case) page, however, the selector box offers two more options:
one to create a new case from scratch, and one to create a case
from a *case* parameter file, which
would then populate values of all parameters found
in the [Case Setup](#case-setup) section.
This box is present on all pages except those in the [Resources](#resource) section.
The case being currently displayed is marked with a small red triangle.

A typical workflow for exploring different scenarios involves starting with a base
case and then duplicating/creating derived scenarios with slight changes in the parameters.
A comparison between the
different resulting outcomes can be found on the [Output Files](#description-output-files) page.
The [Typical Workflow](#typical-workflow) section below
goes through a more specific example.

Owl uses a full year as the standard unit of time. All values are therefore entered and
reported as yearly values. These include wages, income, rates, social security, etc.
Dollar values are typically entered in thousands, unless in tables, where they
are entered and reported in unit dollars.
Graphs report values in thousands, either in nominal value or in today's \\$, as selected.

There are four sections in the menu bar:
[Case Setup](#case-setup), [Single Scenario](#single-scenario),
[Multiple Scenarios](#multiple-scenarios), and [Resources](#resources).
The sections below follow the same logical order.

-------------------------------------------------
### :orange[Case Setup]
This section contains the steps for creating and configuring case scenarios.

#### :material/person_add: Create Case
The **Create Case** page is where every new scenario begins.
It controls the creation of scenarios as the `Case selector` drop-down menu contains
two additional items:
one to create new cases, and the other to create cases from a *case* parameter file.
This page also allows you to duplicate and/or rename scenarios, as well as deleting them.

For creating a scenario from scratch, (first) name(s), marital status, birth year(s),
and life expectancies are required.
Plan starts on Jan 1st of this year and ends on December 31st of the year when all individuals
have passed according to the specified life expectancies.

A typical workflow will involve creating
a base case, and duplicating it a few times with slight changes in its parameters
in order to investigate their effects.
Duplication renames the case by appending a number counter in parenthesis, just as creating
a copy of a file on Windows.
It is recommended to rename each case to reflect the change in parameters.
When duplicating a scenario, make sure to visit all pages in the [Case Setup](#case-setup)
section and verify that all parameters are as intended.
When all cases have successfully run,
results of related cases are compared side-by-side with differences
in the [Output Files](#description-output-files) section.
Related cases are determined by having the same individual's names:
anything else can change between cases.

##### Initializing the life parameters of a scenario
While on the **Create Case** page,
click on the `Case selector` box and choose one of `New case...` or `Upload case file...`.

##### Creating a case from scratch
When starting from `New case...`,
one must provide the year of birth of each spouse(s) and their expected lifespan(s).
For selecting your own numbers, there are plenty of longevity predictors on the Internet.
Pick your favorite:
- [longevityillustrator](https://longevityillustrator.org),
- [livingto100](https://www.livingto100.com/calculator),

or just Google *life expectancy calculator*.

##### Using a *case* file
If `Upload case file...` is selected, a *case* file must be uploaded.
These files end with the *.toml* extension, are human readable (and therefore editable),
and contain all the parameters required to characterize a scenario.
An example is provided
[here](https://github.com/mdlacasse/Owl/main/blob/examples/case_jack+jill.toml?raw=true) and more
can be found in this [directory](https://github.com/mdlacasse/Owl/blob/main/examples/).
Using a *case* file
will populate all the fields in the [Case Setup](#case-setup) section.

A *case* file for the case being developed
can be saved under the [Output Files](#description-output-files) page and
can be reloaded at a later time.
Case parameter files can have any name but when saving from the interface,
their name will start with *case_* followed by the case name.

#### :material/work_history: Wages and Contributions
This page allows to enter an optional Excel file containing future wages and contributions,
and past Roth contributions and conversions.
Alternatively, values can be entered and/or edited directly into the corresponding tables.
Values in these tables are all in nominal values, and in \\$, not thousands (\\$k).
The **Wages and Contributions** table contains 9 columns titled as follows:

|year|anticipated wages|ctrb taxable|ctrb 401k|ctrb Roth 401k|ctrb IRA|ctrb Roth IRA|Roth conv|big-ticket items|
|--|--|--|--|--|--|--|--|--|
|2020 | | | | | | | | |
| ... | | | | | | | | |
|2025 | | | | | | | | |
|2026 | | | | | | | | |
| ... | | | | | | | | |
|20XX | | | | | | | | |

Note that column names are case sensitive and all entries are in lower case.
The easiest way to complete the process of filling this file is either to start from the template
file provided [here](https://github.com/mdlacasse/Owl/blob/main/examples/template.xlsx?raw=true) or
to fill in the values using the user interface, but this last approach does not provide
Excel capabilities for cross-column calculations.

This file goes 5 year back in time in order to capture previous contributions and
conversions to Roth accounts.
Entries in columns others than contributions or conversions to Roth accounts
for past years will be ignored by Owl but can be left there for documentation purposes.
Past contributions and conversions are required for implementing
constraints restricting withdrawals from Roth accounts, thus avoiding
penalties resulting from breaking the 5-year maturation rule.
For that purpose, a retainer on the tax-free account
is put as the sum of all Roth conversions performed
during the last 5 years plus an additional amount for potential
compounded gains. These gains are assumed to be 10% per year for the past
years, and use the predicted returns for future years.
Unlike conversions, contributions can be withdrawn, but a retainer is
put covering the sum of all potential gains resulting from contributions
made over the last 5 years.
However, if funds were converted through a Roth backdoor, this is
still considered as a conversion, and retainer will include contribution amount.
While this approach is somehow more restrictive than
the actual rules, it avoids unnecessary penalties while being a somehow simple approach.
An exact calculation would require to know (and input) the annual rates of return for
the last 5 years and asset allocation ratios, and the same for all future years.
Note that in certain cases, constraints on Roth withdrawals can make a zero bequest impossible
if Roth conversions took place in the 5 years before passing.

Here, year 20XX represents the last row which could be the last year based on the life
expectancy values provided.
While loading an Excel workbook, missing years or empty cells will be filled with zero values,
while years outside the time span of the plan will simply be ignored with the exception
of 5-year back history.

The column *anticipated wages* is the annual amount
(gross minus tax-deferred contributions) that you anticipate to receive from employment
or other sources (e.g., rentals).
This column does not include dividends from your taxable investment accounts,
as they will be calculated based on your assumptions for future return rates.

For the purpose of planning, there is no clear definition of retirement age. There will be a year,
however, from which you will stop having anticipated income, or diminished income due to decreasing your
work load. This transition can be gradual or sudden, and can be explored through these wages
and contributions tables. The only *hard* dates are the years when you intend to receive
a pension or collect social security, and these years are entered elsewhere on the
[Fixed Income](#currency-exchange-fixed-income) page.

Contributions to your savings accounts are marked as *ctrb*. We use 401k as a term that includes
contributions to 403b as well or any other tax-deferred account, with the exception
of IRAs accounts which are treated separately to facilitate data entry.
Contributions to your 401k/403b must also include your employer's
contributions, if any. As these data can be entered in Excel,
one can use the native calculator to enter a percentage
of the anticipated wages for contributions to savings accounts as an easier way to enter the data.
For this purpose, additional columns can coexist in the Excel file
and can be used for tracking other quantities from which
the needed values can be derived. These extra columns will be ignored when the file is processed.

Manual Roth conversions can be specified in the column marked *Roth conv*.
This column is provided to override the Roth conversion optimization in Owl.
When the option `Convert as in contribution file` is toggled
in the [Optimization Parameters](#tune-optimization-parameters) page,
values from the **Wages and Contributions** file will be used and no optimization on Roth conversions
will be performed. This column is provided for flexibility and to allow comparisons
between an optimized solution and your best guesses.

Finally, *big-ticket items* are used for accounting for the sale or purchase of a house, or any
other major expense or money that you would give or receive (e.g., inheritance, or large gifts
to or from you). Therefore, the sign (+/-) of entries in this column is important.
Positive numbers will be considered in the cash flow for that year and the surplus, if any, will be
deposited in the taxable savings accounts. Negative numbers will potentially generate additional
withdrawals and distributions from retirement accounts. This is the only column that can contain
negative numbers: all other column entries must be positive.

When loading an Excel workbook, each individual in the plan must have an associated sheet
for reporting yearly transactions affecting the plan. The association is made by having
the individual's name as the sheet name in the workbook.
Therefore, if preparing your own case using a template, you will need to rename the tabs in the file to
match the names used when creating the plan
(i.e., *Jack* and *Jill* in the example files provided).
If a file was originally associated with a *case* file, a message will remind the user to upload the file.

If values were entered or edited directly in the table,
values can be saved directly in Excel format by clicking
the `Download Wages and Contributions` on the
[Output Files](#description-output-files) page. This allows to rerun the same case at a later time
by reloading the same **Wages and Contributions** file.

#### :material/currency_exchange: Fixed Income
This page is for entering anticipated fixed income from pensions and social security.
Amounts are in thousands (\\$k) in today's \\$. While
social security is always adjusted for inflation, pensions can optionally be
by selecting the corresponding button.
The amounts provided should be as if the benefits start on
January 1st, i.e., before the month when individuals will actually reach the age selected.
This is because Owl uses a full year as time unit and
therefore fixed income starts on the first day of the year during which
the individual reaches the age selected, with the amount provided adjusted for inflation.
If born in November, for example, and planning to take social security at age
67, it is more accurate to provide age 68 as a starting age,
as January 1st of the following year is closer to a November birthday.
The amount would then be for taking social security at age 67 and two months.

A great website for guidance on when to start taking social security is
[opensocialsecurity.com](https://opensocialsecurity.com).
And obviously there is [ssa.gov](https://ssa.gov).

#### :material/savings: Current Assets
This page allows to enter account balances in all savings accounts.
Notice that all amounts are entered in units of \\$1,000, referred to as (\\$k).

Three types of savings accounts are considered and are tracked separately for spouses:
- Taxable savings accounts (e.g., investment accounts, CDs),
- Tax-deferred savings accounts (e.g., 401k, 403b, IRA),
- Tax-free savings accounts (e.g., Roth 401k, Roth IRA).

Account values are assumed to be known at the beginning of the current year,
which is not always possible. For that purpose,
the `Account balance date` has the effect of back projecting the amounts entered
to the beginning of the year using the return rates and allocations
assumed for the first year. If withdrawals contributing to the
net spending were already performed for the current year,
true account balances should be corrected to reflect values as of Jan 1st.

For married couples, the spousal `Beneficiary fractions` associated with each account
can be configured, as well as a surplus deposit fraction. The first one controls
how much is left to the surviving spouse while the second determines
how to split potential surplus budget moneys between the taxable accounts of the spouses.
When the `Beneficiary fractions` are not all 1, it is recommended to deposit all
surplus moneys in the taxable account of the first individual to pass. Otherwise,
the optimizer will find creative solutions that can generate surpluses in order
to maximize the final bequest. Finally, when fractions between accounts are not all equal,
it can take longer to solve (minutes) as these cases trigger the use
of binary variables which involves more complex algorithms.
In some situations, creative transfers from tax-deferred savings accounts to taxable
savings accounts, through surpluses and deposits, can be part of the optimal solution.

Setting a surplus fraction that deposits some or all surpluses in the survivor's account
can also sometimes lead to slow convergence. This is especially noticeable when solving with
varying rates and not so common when using fixed rates.
When using varying rates, it is recommended to set surpluses to be
deposited in the taxable account of first spouse to pass unless exploring specific scenarios.

#### :material/percent: Asset Allocation
This page allows you to select how to partition your assets between 4 investment options,
one equity and three fixed-income securities:
- S&P 500,
- Corporate Bonds Baa,
- 10-year Treasury Notes,
- Cash assets assumed to follow inflation.

When using historical data, the term S&P 500 represents the real index. However,
when selecting non-historical rates, the term can represent any mix of stocks or equities
(domestic, international, emerging, etc.).
The main difference between equities and fixed securities lies
in the tax treatment of gains realized in the taxable account:
equities will be taxed differently than fixed securities.
Cash assets are securities such as Treasury Inflation-Protected Securities (TIPS)
considered to merely track inflation and therefore remain at constant value.

Two choices of asset allocations are possible:
`account` and `individual`. For `account` type, each type
of individual savings account is associated with its own asset allocation ratios.
For `individual`, it is assumed that all savings accounts of a given
individual follow the same allocation ratios.
Allocation ratios can vary over the duration of the plan, starting
from an `initial` allocation ratio at the beginning of the plan
to a `final` allocation ratio at the passing of the individual.
It is assumed that the accounts are regularly
rebalanced to maintain the prescribed allocation ratios.

A gliding function (either `linear` or an `s-curve`) interpolates the values
of the allocation ratios from the `initial` values to the `final` values as the plan progresses in time.
When an `s-curve` is selected, two additional parameters controlling the shape of the transition
will appear, one for the timing of the inflection point measured in years from now,
and the other for the width of the transition, measured in +/- years from the inflection point.

#### :material/monitoring: Rates Selection
This page allows you to select the return rates over the
time span of the plan. All rates are nominal and annual.
There are two major types of rates:
- `Fixed` - staying the same from one year to another:
    - `conservative`,
    - `optimistic`,
    - `historical average` - i.e., average over a range of past years,
    - `user` - rates are provided by the user.
- `Varying` - changing from year to year:
    - `historical` - using a rate sequence which happened in the past,
    - `histochastic` - using stochastic rates derived from statistics over a time range of historical rates,
    - `stochastic` - using stochastic rates created from statistical parameters specified by the user.

These rates are the annual rates of return for each of the assets considered.
The types of asset are described in the previous section.
Rates for the S&P 500 equities include dividends.
A roundup of expert opinions on stock and bond return forecasts for the next decade can be found
[here](https://www.morningstar.com/portfolios/experts-forecast-stock-bond-returns-2025-edition) for 2025.

This page also includes some adjustments related to future tax rates. One is the anticipated
tax rate that heirs will pay on the tax-deferred portion of the bequest. Another setting is related
to the year when the Tax Cut and Job Act rates is anticipated to be expired.

#### :material/tune: Optimization Parameters
This page allows you to select the objective function to optimize.
One can choose between maximizing the net spending amount subject to the constraint
of a desired bequest, or maximizing a bequest, subject to the constraint of providing
a desired net spending amount.
As one of the two choices (net spending or bequest) is selected as the value to maximize,
the other becomes a constraint to obey.

The maximum amount for Roth conversions and which spouse can execute them is configurable.
Roth conversions are optimized for reducing taxes and maximizing the selected objective function,
unless the `Convert from contributions file`
button is toggled, in which case Roth conversions will not be optimized,
but will rather be performed according to
the *Roth conv* column on the
[Wages and Contributions](#work-history-wages-and-contributions) page.
A year from which Roth conversions can begin to be considered can also be selected:
no Roth conversions will be allowed before the year specified.

A self-consistent loop is used to compute values that are difficult
to integrate into a linear program. These includes the net investment income tax (NIIT),
the rate on capital gains (0, 15, or 20%), the phase out of the additional exemption for seniors,
and potentially the Medicare and IRMAA premiums.
Turning off the self-consistent loop will default all these values to zero.

An additional setting allows to turn off mutually exclusive operations,
such as Roth conversions and withdrawals from the tax-free account.
Enabling these mutually exclusive constraints avoids both these situations.
Surprinsingly, dropping these constraints can lead to slightly different optimal points
for reasons that escape me.

Medicare premiums start automatically in the year each individual reach age 65.
Calculations of Medicare and IRMAA can be turned off, calculated through
the self-consistent loop, or be integrated into the optimization.
Optimizing (as opposed to calculating) Medicare and IRMAA can sometimes be much slower given
the additional number of binary variables involved. This option should be used
for a single case, and compared with results obtained from self-consistent calculations.
Medicare optimization should not be used when running multiple scenarios such
as in Monte Carlo simulations.

If the current age of any individual in the plan makes them eligible
for Medicare within the next two years,
additional cells will appear for entering the MAGI's for the
past 1 or 2 years, whichever applies. Values default to zero.
These numbers are needed to calculate the Income-Related Monthly Adjusted Amounts (IRMAA).

Different mixed-integer linear programming solvers can be selected.
This option is mostly for verification purposes.
All solvers tested
(HiGHS, COIN-OR Branch-and-Cut solver through PuLP, HiGHS through PuLP, and MOSEK)
provided very similar results.
Due to the mixed-integer formulation, solver performance is sometimes unpredictable.
In general, CBC tends to be slower, partly because of the algorithm,
and partly because it solves the problem through a model description saved in
a temporary file requiring I/O.
In most cases, selecting `HiGHS` will provide great results in the shortest time.

The time profile modulating the net spending amount
can be selected to either be `flat` or follow a `smile` shape.
The *smile* shape has three configurable parameters: a `dip` percentage
a linear `increase` (or decrease if negative), over the time period (apart from inflation),
and a time `delay`, in years from today, before the non-flat behavior starts to act.
Values default to 15%, 12%, and 0 year respectively, but they are fully configurable
for experimentation and to fit your anticipated lifestyle.

A `slack` variable can also be adjusted. This variable allows the net spending to deviate from
the desired profile in order to maximize the objective. This is provided mostly for educational purpose
as maximizing the total net spending will involve leaving the savings invested for as long as possible,
and therefore this will favor smaller spending early in the plan and larger towards the end.
This tension between maximizing a dollar amount and the utility of money then becomes evident.
While the health of the individuals and therefore the utility of money is higher at the beginning
of retirement, maximizing the total spending or bequest will pull in the opposite direction.

For married couples, the survivor's
net spending percentage is also configurable. A value of 60% is typically used.
The selected profile multiplies
the net spending *basis* which sets the resulting spending
amounts over the duration of the plan.
Notice that *smile* curves are re-scaled to have the same total spending as flat curves:
for that reason they do not start at 1.

--------------------------------------------------------------------------------------
### :orange[Single Scenario]

#### :material/stacked_line_chart: Graphs
This page displays various plots from a single scenario based on the selections made
in the [Case Setup](#case-setup) section.
This simulation uses a single instance of a series of rates, either fixed or varying,
as selected in the [Case Setup](#case-setup) section.
The outcome is optimized according to the chosen parameters: either maximize the
net spending, of maximize the bequest under the constraint of a net spending amount.
Various plots show the results, which can be displayed in today's \\$ or
in nominal value.

A button allows to re-run the case which would generate a different result
if the chosen rates are `histochastic` or `stochastic`. Each graph can be seen
in full screen, and are interactive when using the `plotly` library.
Graphs can be drawn using the `matplotlib` or `plotly` libraries as
selected in the [Settings](#settings-settings) section described below.

#### :material/data_table: Worksheets
This page shows the various worksheets containing annual transactions
and savings account balances in nominal \\$.
Savings balances are values at the beginning of the year, while other quantities
are for the full year.
Each table can be downloaded separately in csv format, or all tables can be downloaded
jointly as a single Excel workbook by clicking on the `Download Worksheets` on the
[Output Files](#description-output-files) page.
Note that all values here (worksheets and workbook) are in \\$, not in thousands.
The first few lines of the **Sources** worksheets are the most important
as these lines are the only ones that are actionable in the near term.

#### :material/description: Output Files
This page allows to compare cases and save files for future use.
First, it shows a synopsis of the computed scenario by
displaying sums of income, bequest, and spending values over the duration of the plan.
Note that all tables are scrollable and can be seen in full-screen mode.
If multiple cases were configured and run (most likely through duplication and
modifying the parameters), they will be compared in that panel provided they were made
for the same individuals and years spans. Column on the left shows the values for the selected case
while those on the right shows the differences.
The contents of the synopsis can be downloaded as a plain text file by
clicking the button below it.
An additional button allows to rerun all cases,
ensuring that the table provides an accurate comparison
of the parameters selected for each case.

Another section called **Excel Workbooks** allows
to save the contents of the tables on corresponding pages as a single Excel workbook.
The `Download Wages and Contributions file` will save the data displayed on the
[Wages and Contributions](#work-history-wages-and-contributions) page while the
`Download Worksheets` will save all tables displayed
on the [Worksheets](#data-table-worksheets) page as a single Excel file.

Similarly, all parameters used to generate the case are collected in *toml* format and displayed.
The `Download case parameter file...` button allows to save the parameters of the selected scenario
to a *case* file for future use.

With the case parameter file and the **Wages and Contributions** worksheet,
the same case can be reproduced at a later time by uploading
them through the widgets on the [Create Case](#person-add-create-case)
and [Wages and Contributions](#work-history-wages-and-contributions) pages,
respectively.

--------------------------------------------------------------------------------------
### :orange[Multiple Scenarios]
There are two different ways to run multiple scenarios and generate a histogram
of results.

#### :material/history: Historical Range
This page allows the user to run multiple simulations over a range of historical years.
Each simulation assumes that the rates follow a sequence that happened in the past,
starting from each selected year in the past, and then offset by one year, and so on.
A histogram of results and a success rate is displayed at the end of the run.
$N$ is the number of runs that could fit in the year range selected,
$P$ the probability of success,
$\\bar{x}$ is the resulting average, and $M$ is the median.

If the `Beneficiary fractions` are not all unity, two histograms will be displayed:
one for the partial bequest left at the passing of the first spouse
and the other for the distribution of values of the objective being optimized,
either maximum net spending or maximum bequest left at the passing
of the surviving spouse.

#### :material/finance: Monte Carlo
This page runs a Monte Carlo simulation using time sequences of
annual rates of return that are generated
using statistical methods. At the end of the run,
a histogram is shown, with a probability of success.

The mean outcome $\\bar{x}$ and the median $M$ are provided in the graph, as are the number
of cases $N$ and the probability of success $P$, which is the percentage of cases that succeeded.
Cases that failed are termed infeasible, as the optimizer could not find
values that could satisty all constraints.

As is the case for [Historical Range](#history-historical-range),
if the `Beneficiary fractions` are not all unity, two histograms will also be displayed:
one for the partial bequest left at the passing of the first spouse
and the other for the distribution of values of the objective being optimized,
either maximum net spending or maximum bequest left at the passing
of the surviving spouse.

Linear programming solutions are more expensive than event-driven forward simulators.
Therefore, when considering Monte Carlo simulations, consider:
- Turning off Medicare calculations,
- Installing Owl and running on your local computer as it can sometimes be
faster than running on the Streamlit host, depending on your hardware.
Moreover, the community server has a
CPU time limit that will stop a session after the quota is reached.
Most likely, this will not happen unless you devise unusually long Monte Carlo runs.

--------------------------------------------------------------------------------------
### :orange[Resources]
#### :material/rocket_launch: Quick Start
This page is the landing page of the application.
It shows new users how to quickly get started by using an example case file.

#### :material/help: Documentation
These very pages.

#### :material/settings: Settings
This page allows to select different backends for plotting the graphs.
The `plotly` package is currently the default as the graphs generated are interactive
while `matplotlib` graphs are not.
Plots generated by `matplotlib`, however, have a more traditional look
and form the most mature part of the code.

When using `plotly`, users can zoom and pan, and toggle
traces by clicking on the associated legend. Double clicking on a legend item
selects only this item. Double-clicking on a disabled item restores all traces.
A copy of the plot can also be saved as a *png* file. Both `matplotlib` and `plotly`
graphs can be seen in full-screen mode.

The position of the menubar can be selected to be at the top or as a sidebar.
The sidebar menu can also be collapsed if needed.
Default behavior is to have the menubar at the top, unless on a mobile device.

###### Streamlit App and Theme
If you are accessing Owl remotely on the Streamlit Community Server through the Chrome browser,
the Chrome performance manager might disable hidden or inactive tabs.
This could cause your Owl session to inadvertently reset when idling for too long,
and losing the state of the calculator.
The best way to avoid this situation is to run the web page through the Streamlit app on your device.
This is done by clicking the '+" icon shown at the right end of the browser URL bar,
showing *App available: Install Streamlit*.
The app provides more screen space as it doesn't have a navigation bar.
On a mobile device, saving the page to the home screen will achieve the same result.
A similar timing problem can happen if your simulations
(Monte Carlo) are extremely long.
The Streamlit community server has a hard resource limit
on CPU time that might stop your calculations before completion.
I could successfully run 1,000 simulations using the Streamlit app while
being hosted on the Streamlit community server.
However, if you are contemplating running Monte Carlo simulations
with thousands of cases routinely,
you should definitely consider installing and running Owl
locally on your computer, either natively or through a container.
See instructions on the GitHub repository for how to proceed.

If not using the Streamlit app, going full screen by hitting F11 in your browser
can also greatly improve the visualization of graphs and worksheets.

Owl's default theme is the *Dark* mode but a *Light* theme is also available by
clicking on the three vertical dots located on the upper right of the app
and selecting the **Settings** option.

#### :material/error: Logs
Messages coming from the underlying Owl calculation engine are displayed on this page.
This page is mainly used for debugging purposes.

#### :material/info: About Owl
Credits and disclaimers.

--------------------------------------------------------------------------------------
### :orange[Tips]
#### :material/lightbulb_2: Advice on Optimization and Roth Conversions
Owl does not explicitly optimize for Medicare costs. While this approach
is possible and was tested,
it is not practical due to the unpredictable computing time
required by the additional number of binary variables involved.
The current approach is to add
Medicare costs after the optimization step.
As a result, the suggested Roth conversions can sometimes lead to
smaller net spending or bequest than when no conversions are made.
This is due to higher Medicare costs triggered by the increased MAGI
resulting from Roth conversions
which are factored in during the optimization step.
This is why one should **always** run comparisons between cases with and without Roth conversions.
Also keep in mind that these cases only consider current assumptions and obviously
do not take into account future income tax rate increases.

#### :material/rule_settings: Typical Workflow
A typical workflow would look like the following:
1) Create a base case representing your basic scenario;
2) Duplicate the base case and modify the parameter you want to investigate;
3) Repeat 2) with other end-member values of the parameter you would like to consider;
4) Run all cases and compare them on the [Output Files](#description-output-files) page.

To make it more concrete, here is an example
where one would like to investigate the effects of Roth conversions
on total net spending.
1) Create a case called, say, *May 2025 - Base case*.
Fill in all parameters representing your goals and situation.
Upload file or fill-in values for Wages and Contributions.
Let's say this case allows for Roth conversions up to \\$100k.
2) Duplicate the base case, call it *May 2025 - No Roth conversions* and
set maximum Roth conversions to 0.
3) Duplicate the base case again, call it *May 2025 - No Roth limit* and
set maximum Roth conversions to a very large number, say, \\$800k.
4) Compare all cases on the [Output Files](#description-output-files) page.

As mentionned above, the most actionable information is located on the first few lines
of the **Sources** tables on the [Worksheets](#data-table-worksheets) pages.
This is where withdrawals and conversions are displayed for this year and the next few years.

#### :material/mindfulness: Scope of Use
In general, computer modeling is not about predicting the future,
but rather about exploring possibilities.
Owl has been designed to allow users to explore multiple parameters
and their effects on their future financial situation while avoiding the trap of overmodeling.
Overmodeling happens when a computer model has a level of detail
far beyond the uncertainty of the problem, which can lead some users to assume
an unjustified level of certainty in the results.

The main goal of retirement financial planning is to characterize
uncertainty and translate this uncertainty into actionable, near-term decisions.
Therefore, as a deliberate choice in design, state tax and complex federal tax rules were
not included into Owl. As can be seen from this
[graph](https://marottaonmoney.com/wp-content/uploads/2023/04/Historical-Effective-Rates-Through-2023.jpg),
income tax rates have varied a lot over the last century. Assuming that current rates will
stay fixed for the next 30 years is just unrealistic. But the best assumption
we have is to predict rates from what we currently know,
or by projecting historical data into the future.
This approach allows us to frame the problem within a range of likely scenarios.

Users must always keep in mind that retirement financial planning tools have inherent limitations
and that common sense will always be the best ally. Plans need to be revisited
as new information is obtained, allowing to make regular corrections to
the best estimates.
Understanding the limitations of any retirement financial planning tool is absolutely critical
to interpreting the results they provide.
""")



"""

Owl/abcapi
---

A retirement planner using linear programming optimization.

See companion document for a complete explanation and description
of all variables and parameters.

This file contains basic functions to build a constraint matrix and
objective function line by line. This is used to abstract the
building of the constraint matrix in order to be able to use various
solvers for comparison.

This approach has been successful with the MOSEK and the HiGHS solvers.
A for matrix, B for bounds, C for constraints. Thus the name ABCAPI.

Copyright &copy; 2024 - Martin-D. Lacasse

Disclaimers: This code is for educational purposes only and does not constitute financial advice.

"""

import numpy as np


class Row(object):
    """
    Solver-neutral API to accomodate Mosek/HiGHS.
    A Row represent a row in matrix A.
    """

    def __init__(self, nvars):
        """
        Constructor requires the number of decision variables.
        """
        self.nvars = nvars
        self.ind = []
        self.val = []

    def addElem(self, ind, val):
        """
        Add an element at index ``ind`` of value ``val`` to the row.
        """
        if not (0 <= ind < self.nvars):
            raise ValueError(f"Index {ind} out of range.")
        self.ind.append(ind)
        self.val.append(val)

    def addElemDic(self, rowDic=None):
        """
        Add elements at indices provided by a dictionary.
        """
        rowDic = {} if rowDic is None else rowDic
        for key in rowDic:
            self.addElem(key, rowDic[key])
        return self


class ConstraintMatrix(object):
    """
    Solver-neutral API for expressing constraints.
    """

    def __init__(self, nvars):
        """
        Constructor only requires the number of decision variables.
        """
        self.ncons = 0
        self.nvars = nvars
        self.Aind = []
        self.Aval = []
        self.lb = []
        self.ub = []
        self.key = []

    def newRow(self, rowDic=None):
        """
        Create a new row and populate its elements using the dictionary provided.
        Return the row created.
        """
        rowDic = {} if rowDic is None else rowDic
        row = Row(self.nvars)
        row.addElemDic(rowDic)
        return row

    def addRow(self, row, lb, ub):
        """
        Add row ``row`` to the constraint matrix with the lower ``lb`` and
        upper bound ``ub`` provided.
        """
        self.Aind.append(row.ind)
        self.Aval.append(row.val)
        self.lb.append(lb)
        self.ub.append(ub)
        if lb == ub:
            self.key.append("fx")
        elif ub == np.inf and lb == -np.inf:
            self.key.append("fr")
        elif ub == np.inf:
            self.key.append("lo")
        elif lb == -np.inf:
            self.key.append("up")
        else:
            self.key.append("ra")
        self.ncons += 1

    def addNewRow(self, rowDic, lb, ub):
        """
        Create and add a new row to the constraint matrix with the lower ``lb`` and
        upper bound ``ub`` provided. Row's elements are populated from the provided
        dictionary ``rowDic``.
        """
        row = self.newRow(rowDic)
        self.addRow(row, lb, ub)

    def keys(self):
        """
        Return list of keys for each row used by MOSEK.
        """
        return self.key

    def lists(self):
        """
        Return lists of indices and values for MOSEK sparse representation.
        """
        return self.Aind, self.Aval, self.lb, self.ub

    def arrays(self):
        """
        Return full arrays for Scipy/HiGHS.
        """
        Alu = np.zeros((self.ncons, self.nvars))
        lb = np.array(self.lb)
        ub = np.array(self.ub)
        for ii in range(self.ncons):
            ind = self.Aind[ii]
            val = self.Aval[ii]
            for jj in range(len(ind)):
                Alu[ii, ind[jj]] = val[jj]

        return Alu, lb, ub


class Bounds(object):
    """
    Solver-neutral API for bounds on variables.
    """

    def __init__(self, nvars, nbins):
        self.nvars = nvars
        self.nbins = nbins
        self.ind = []
        self.lb = []
        self.ub = []
        self.key = []
        self.integrality = []
        for ii in range(nvars-nbins, nvars):
            self.setBinary(ii)

    def setBinary(self, ii):
        if not (0 <= ii < self.nvars):
            raise ValueError(f"Index {ii} out of range.")
        self.ind.append(ii)
        self.lb.append(0)
        self.ub.append(1)
        self.key.append("ra")
        self.integrality.append(ii)

    def setRange(self, ii, lb, ub):
        if not (0 <= ii < self.nvars):
            raise ValueError(f"Index {ii} out of range.")
        if lb > ub:
            raise ValueError(f"Lower bound {lb} > upper bound {ub}.")
        self.ind.append(ii)
        self.lb.append(lb)
        self.ub.append(ub)
        if lb == ub:
            self.key.append("fx")
        elif ub == np.inf and lb == -np.inf:
            self.key.append("fr")
        elif ub == np.inf:
            self.key.append("lo")
        elif lb == -np.inf:
            self.key.append("up")
        else:
            self.key.append("ra")

    def keys(self):
        keys = ["lo"] * self.nvars
        for ii in range(len(self.ind)):
            keys[self.ind[ii]] = self.key[ii]

        return keys

    def arrays(self):
        lb = np.zeros(self.nvars)
        ub = np.ones(self.nvars) * np.inf
        for ii in range(len(self.ind)):
            lb[self.ind[ii]] = self.lb[ii]
            ub[self.ind[ii]] = self.ub[ii]

        return lb, ub

    def integralityArray(self):
        integrality = np.zeros(self.nvars, dtype=int)
        for ii in range(len(self.integrality)):
            integrality[self.integrality[ii]] = 1

        return integrality

    def integralityList(self):
        return self.integrality


class Objective(object):
    """
    Solver-neutral objective function.
    """

    def __init__(self, nvars):
        self.nvars = nvars
        self.ind = []
        self.val = []

    def setElem(self, ind, val):
        if not (0 <= ind < self.nvars):
            raise ValueError(f"Index {ind} out of range.")
        self.ind.append(ind)
        self.val.append(val)

    def arrays(self):
        """
        Return an array for scipy/HiGHS dense representation.
        """
        c = np.zeros(self.nvars)
        for ii in range(len(self.ind)):
            c[self.ind[ii]] = self.val[ii]

        return c

    def lists(self):
        """
        Return lists for Mosek sparse representation.
        """
        return self.ind, self.val



"""
A simple object to display progress in Streamlit.

"""

import streamlit as st


class Progress:
    def __init__(self, mylog):
        self.mylog = mylog
        self.counter = 0
        self.clocks = []
        for i in range(1, 13):
            self.clocks.extend([f":clock{i}:", f":clock{i}30:"])
        self.txt = "Calculations in progress. Please wait. "

    def msg(self):
        self.counter += 1
        return self.txt + self.clocks[(self.counter - 1) % 24]

    def start(self):
        self.bar = st.progress(0, self.msg())

    def show(self, x):
        self.bar.progress(x, self.msg())

    def finish(self):
        self.bar.empty()



"""
Matplotlib implementation of plot backend.

Copyright &copy; 2025 - Martin-D. Lacasse

Disclaimers: This code is for educatonal purposes only and does not constitute financial advice.

"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.ticker as tk
import io
import os

os.environ["JUPYTER_PLATFORM_DIRS"] = "1"

import seaborn as sbn           # Noqa: E402

from .base import PlotBackend   # Noqa: E402
from .. import utils as u       # Noqa: E402


class MatplotlibBackend(PlotBackend):
    """Matplotlib implementation of plot backend."""

    def __init__(self):
        """Initialize the matplotlib backend."""
        self.set_plot_style()

    def jupyter_renderer(self, fig):
        pass
        # plt.show()

    def set_plot_style(self):
        """Set the style for all matplotlib plots."""
        plt.rcParams.update({'figure.autolayout': True})
        plt.rcParams.update({'figure.figsize': (6, 4)})
        plt.rcParams.update({'axes.grid': True})
        plt.rcParams.update({'axes.grid.which': 'both'})

    def _line_income_plot(self, x, series, style, title, yformat=r"\$k"):
        """Core line plotter function."""
        fig, ax = plt.subplots()

        for sname in series:
            ax.plot(x, series[sname], label=sname, ls=style[sname])

        ax.legend(loc="upper left", reverse=True, fontsize=8, framealpha=0.3)
        ax.set_title(title)
        ax.set_xlabel("year")
        ax.set_ylabel(yformat)
        ax.xaxis.set_major_locator(tk.MaxNLocator(integer=True))
        if "k" in yformat:
            ax.get_yaxis().set_major_formatter(tk.FuncFormatter(lambda x, p: format(int(x / 1000), ",")))
            # Give range to y values in unindexed flat profiles.
            ymin, ymax = ax.get_ylim()
            if ymax - ymin < 5000:
                ax.set_ylim((ymin * 0.95, ymax * 1.05))

        return fig, ax

    def _stack_plot(self, x, inames, title, irange, series, snames, location, yformat=r"\$k"):
        """Core function for stacked plots."""
        nonzeroSeries = {}
        for sname in snames:
            for i in irange:
                tmp = series[sname][i]
                if sum(tmp) > 1.0:
                    nonzeroSeries[sname + " " + inames[i]] = tmp

        if len(nonzeroSeries) == 0:
            return None, None

        fig, ax = plt.subplots()

        ax.stackplot(x, nonzeroSeries.values(), labels=nonzeroSeries.keys(), alpha=0.6)
        ax.legend(loc=location, reverse=True, fontsize=8, ncol=2, framealpha=0.5)
        ax.set_title(title)
        ax.set_xlabel("year")
        ax.xaxis.set_major_locator(tk.MaxNLocator(integer=True))
        if "k" in yformat:
            ax.set_ylabel(yformat)
            ax.get_yaxis().set_major_formatter(tk.FuncFormatter(lambda x, p: format(int(x / 1000), ",")))
        elif yformat == "percent":
            ax.set_ylabel("%")
            ax.get_yaxis().set_major_formatter(tk.FuncFormatter(lambda x, p: format(int(100 * x), ",")))
        else:
            raise RuntimeError(f"Unknown yformat: {yformat}.")

        return fig, ax

    def plot_histogram_results(self, objective, df, N, year_n, n_d=None, N_i=1, phi_j=None):
        """Show a histogram of values from historical data or Monte Carlo simulations."""
        description = io.StringIO()

        pSuccess = u.pc(len(df) / N)
        print(f"Success rate: {pSuccess} on {N} samples.", file=description)
        title = f"$N$ = {N}, $P$ = {pSuccess}"
        means = df.mean(axis=0, numeric_only=True)
        medians = df.median(axis=0, numeric_only=True)

        my = 2 * [year_n[-1]]
        if N_i == 2 and n_d is not None and n_d < len(year_n):
            my[0] = year_n[n_d - 1]

        # Don't show partial bequest of zero if spouse is full beneficiary,
        # or if solution led to empty accounts at the end of first spouse's life.
        if (phi_j is not None and np.all((1 - phi_j) < 0.01)) or medians.iloc[0] < 1:
            if medians.iloc[0] < 1:
                print(f"Optimized solutions all have null partial bequest in year {my[0]}.", file=description)
            df.drop("partial", axis=1, inplace=True)
            means = df.mean(axis=0, numeric_only=True)
            medians = df.median(axis=0, numeric_only=True)
            my[0] = my[1]

        nfields = len(means)

        df /= 1000
        if len(df) > 0:
            thisyear = year_n[0]
            if objective == "maxBequest":
                fig, axes = plt.subplots()
                # Show both partial and final bequests in the same histogram.
                sbn.histplot(df, multiple="dodge", kde=True, ax=axes)
                legend = []
                # Don't know why but legend is reversed from df.
                for q in range(nfields - 1, -1, -1):
                    dmedian = u.d(medians.iloc[q], latex=True)
                    dmean = u.d(means.iloc[q], latex=True)
                    legend.append(f"{my[q]}: $M$: {dmedian}, $\\bar{{x}}$: {dmean}")
                plt.legend(legend, shadow=True)
                plt.xlabel(f"{thisyear} $k")
                plt.title(objective)
                leads = [f"partial {my[0]}", f"  final {my[1]}"]
                leads = leads if nfields == 2 else leads[1:]
            elif nfields == 2:
                # Show partial bequest and net spending as two separate histograms.
                fig, axes = plt.subplots(1, 2, figsize=(10, 5))
                cols = ["partial", objective]
                leads = [f"partial {my[0]}", objective]
                for q, col in enumerate(cols):
                    sbn.histplot(df[col], kde=True, ax=axes[q])
                    dmedian = u.d(medians.iloc[q], latex=True)
                    dmean = u.d(means.iloc[q], latex=True)
                    legend = [f"$M$: {dmedian}, $\\bar{{x}}$: {dmean}"]
                    axes[q].set_label(legend)
                    axes[q].legend(labels=legend)
                    axes[q].set_title(leads[q])
                    axes[q].set_xlabel(f"{thisyear} $k")
            else:
                # Show net spending as single histogram.
                fig, axes = plt.subplots()
                sbn.histplot(df[objective], kde=True, ax=axes)
                dmedian = u.d(medians.iloc[0], latex=True)
                dmean = u.d(means.iloc[0], latex=True)
                legend = [f"$M$: {dmedian}, $\\bar{{x}}$: {dmean}"]
                plt.legend(legend, shadow=True)
                plt.xlabel(f"{thisyear} $k")
                plt.title(objective)
                leads = [objective]

            plt.suptitle(title)

            for q in range(nfields):
                print(f"{leads[q]:>12}: Median ({thisyear} $): {u.d(medians.iloc[q])}", file=description)
                print(f"{leads[q]:>12}:   Mean ({thisyear} $): {u.d(means.iloc[q])}", file=description)
                mmin = 1000 * df.iloc[:, q].min()
                mmax = 1000 * df.iloc[:, q].max()
                print(f"{leads[q]:>12}:           Range: {u.d(mmin)} - {u.d(mmax)}", file=description)
                nzeros = len(df.iloc[:, q][df.iloc[:, q] < 0.001])
                print(f"{leads[q]:>12}:    N zero solns: {nzeros}", file=description)

            return fig, description

        return None, description

    def plot_rates_correlations(self, name, tau_kn, N_n, rate_method, rate_frm=None, rate_to=None,
                                tag="", share_range=False):
        """Plot correlations between various rates."""
        rate_names = [
            "S&P500 (incl. div.)",
            "Baa Corp. Bonds",
            "10-y T-Notes",
            "Inflation",
        ]

        df = pd.DataFrame()
        for k, name in enumerate(rate_names):
            data = 100 * tau_kn[k]
            df[name] = data

        g = sbn.PairGrid(df, diag_sharey=False, height=1.8, aspect=1)
        if share_range:
            minval = df.min().min() - 5
            maxval = df.max().max() + 5
            g.set(xlim=(minval, maxval), ylim=(minval, maxval))
        g.map_upper(sbn.scatterplot)
        g.map_lower(sbn.kdeplot)
        g.map_diag(sbn.histplot, color="orange")

        # Put zero axes on off-diagonal plots.
        imod = len(rate_names) + 1
        for i, ax in enumerate(g.axes.flat):
            ax.axvline(x=0, color="grey", linewidth=1, linestyle=":")
            if i % imod != 0:
                ax.axhline(y=0, color="grey", linewidth=1, linestyle=":")

        title = name + "\n"
        title += f"Rates Correlations (N={N_n}) {rate_method}"
        if rate_method in ["historical", "histochastic"]:
            title += f" ({rate_frm}-{rate_to})"

        if tag != "":
            title += " - " + tag

        g.figure.suptitle(title, y=1.08)
        return g.figure

    def plot_rates(self, name, tau_kn, year_n, N_k,
                   rate_method, rate_frm=None, rate_to=None, tag=""):
        """Plot rate values used over the time horizon."""
        fig, ax = plt.subplots()
        title = name + "\nReturn & Inflation Rates (" + str(rate_method)
        if rate_method in ["historical", "histochastic", "historical average"]:
            title += f" {rate_frm}-{rate_to}"
        title += ")"

        if tag != "":
            title += " - " + tag

        rate_name = [
            "S&P500 (incl. div.)",
            "Baa Corp. Bonds",
            "10-y T-Notes",
            "Inflation",
        ]
        ltype = ["-", "-.", ":", "--"]

        for k in range(N_k):
            data = 100 * tau_kn[k]
            # Use ddof=1 to match pandas' statistical calculations from numpy.
            label = (
                rate_name[k] + " <" + "{:.1f}".format(np.mean(data)) + " +/- {:.1f}".format(np.std(data, ddof=1)) + "%>"
            )
            ax.plot(year_n, data, label=label, ls=ltype[k % N_k])

        ax.xaxis.set_major_locator(tk.MaxNLocator(integer=True))
        ax.legend(loc="best", reverse=False, fontsize=8, framealpha=0.7)
        ax.set_title(title)
        ax.set_xlabel("year")
        ax.set_ylabel("%")

        return fig

    def plot_rates_distributions(self, frm, to, SP500, BondsBaa, TNotes, Inflation, FROM):
        """Plot histograms of the rates distributions."""
        title = f"Rates from {frm} to {to}"
        # Bring year values to indices.
        frm -= FROM
        to -= FROM

        nbins = int((to - frm) / 4)
        fig, ax = plt.subplots(1, 4, sharey=True, sharex=True, tight_layout=True)

        dat0 = np.array(SP500[frm:to])
        dat1 = np.array(BondsBaa[frm:to])
        dat2 = np.array(TNotes[frm:to])
        dat3 = np.array(Inflation[frm:to])

        fig.suptitle(title)
        ax[0].set_title("S&P500")
        label = "<>: " + u.pc(np.mean(dat0), 2, 1)
        ax[0].hist(dat0, bins=nbins, label=label)
        ax[0].legend(loc="upper left", fontsize=8, framealpha=0.7)

        ax[1].set_title("BondsBaa")
        label = "<>: " + u.pc(np.mean(dat1), 2, 1)
        ax[1].hist(dat1, bins=nbins, label=label)
        ax[1].legend(loc="upper left", fontsize=8, framealpha=0.7)

        ax[2].set_title("TNotes")
        label = "<>: " + u.pc(np.mean(dat2), 2, 1)
        ax[2].hist(dat2, bins=nbins, label=label)
        ax[2].legend(loc="upper left", fontsize=8, framealpha=0.7)

        ax[3].set_title("Inflation")
        label = "<>: " + u.pc(np.mean(dat3), 2, 1)
        ax[3].hist(dat3, bins=nbins, label=label)
        ax[3].legend(loc="upper left", fontsize=8, framealpha=0.7)

        return fig

    def plot_gross_income(self, year_n, G_n, gamma_n, value, title, tax_brackets):
        """Plot gross income over time."""
        style = {"taxable income": "-"}
        if value == "nominal":
            series = {"taxable income": G_n}
            yformat = r"\$k (nominal)"
            infladjust = gamma_n[:-1]
        else:
            series = {"taxable income": G_n / gamma_n[:-1]}
            yformat = r"\$k (" + str(year_n[0]) + r"\$)"
            infladjust = 1
        fig, ax = self._line_income_plot(year_n, series, style, title, yformat)
        # Overlay tax brackets
        for key in tax_brackets:
            data_adj = tax_brackets[key] * infladjust
            ax.plot(year_n, data_adj, label=key, ls=":")
        ax.grid(visible=True, which='both')
        ax.legend(loc="upper left", reverse=True, fontsize=8, framealpha=0.3)

        return fig

    def plot_profile(self, year_n, xi_n, title, inames):
        """Plot profile over time."""
        style = {"profile": "-"}
        series = {"profile": xi_n}

        return self._line_income_plot(year_n, series, style, title, yformat=r"$\xi$")[0]

    def plot_net_spending(self, year_n, g_n, xi_n, xiBar_n, gamma_n, value, title, inames):
        """Plot net spending over time."""
        style = {"net": "-", "target": ":"}
        if value == "nominal":
            series = {"net": g_n, "target": (g_n[0] / xi_n[0]) * xiBar_n}
            yformat = r"\$k (nominal)"
        else:
            series = {"net": g_n / gamma_n[:-1], "target": (g_n[0] / xi_n[0]) * xi_n}
            yformat = r"\$k (" + str(year_n[0]) + r"\$)"

        return self._line_income_plot(year_n, series, style, title, yformat)[0]

    def plot_asset_composition(self, year_n, inames, b_ijkn, gamma_n, value, name, tag):
        """Plot asset composition over time."""
        if value == "nominal":
            yformat = r"\$k (nominal)"
            infladjust = 1
        else:
            yformat = r"\$k (" + str(year_n[0]) + r"\$)"
            infladjust = gamma_n
        years_n = np.array(year_n)
        years_n = np.append(years_n, [years_n[-1] + 1])
        y2stack = {}
        jDic = {"taxable": 0, "tax-deferred": 1, "tax-free": 2}
        kDic = {"stocks": 0, "C bonds": 1, "T notes": 2, "common": 3}
        figures = []
        for jkey in jDic:
            stackNames = []
            for kkey in kDic:
                namek = kkey + " / " + jkey
                stackNames.append(namek)
                y2stack[namek] = np.zeros((len(inames), len(years_n)))
                for i in range(len(inames)):
                    y2stack[namek][i][:] = b_ijkn[i][jDic[jkey]][kDic[kkey]][:] / infladjust
            title = name + "\nAsset Composition - " + jkey
            if tag:
                title += " - " + tag
            fig, ax = self._stack_plot(years_n, inames, title, range(len(inames)),
                                       y2stack, stackNames, "upper left", yformat)
            figures.append(fig)

        return figures

    def plot_allocations(self, year_n, inames, alpha_ijkn, ARCoord, title):
        """Plot allocations over time."""
        count = len(inames)
        if ARCoord == "spouses":
            acList = [ARCoord]
            count = 1
        elif ARCoord == "individual":
            acList = [ARCoord]
        elif ARCoord == "account":
            acList = ["taxable", "tax-deferred", "tax-free"]
        else:
            raise ValueError(f"Unknown coordination {ARCoord}.")
        figures = []
        assetDic = {"stocks": 0, "C bonds": 1, "T notes": 2, "common": 3}
        blank = ["", ""]
        for i in range(count):
            y2stack = {}
            for acType in acList:
                stackNames = []
                for key in assetDic:
                    # aname = key + " / " + acType
                    aname = key
                    stackNames.append(aname)
                    y2stack[aname] = np.zeros((count, len(year_n)))
                    y2stack[aname][i][:] = alpha_ijkn[i, acList.index(acType), assetDic[key], : len(year_n)]
                t = title + f" - {acType} {inames[i]}"
                fig, ax = self._stack_plot(year_n, blank, t, [i], y2stack, stackNames, "upper left", "percent")
                figures.append(fig)

        return figures

    def plot_accounts(self, year_n, savings_in, gamma_n, value, title, inames):
        """Plot accounts over time."""
        stypes = list(savings_in.keys())
        year_n_full = np.append(year_n, [year_n[-1] + 1])
        if value == "nominal":
            yformat = r"\$k (nominal)"
            savings = savings_in
        else:
            yformat = r"\$k (" + str(year_n[0]) + r"\$)"
            savings = {k: v / gamma_n for k, v in savings_in.items()}
        fig, ax = self._stack_plot(year_n_full, inames, title, range(len(inames)),
                                   savings, stypes, "upper left", yformat)

        return fig

    def plot_sources(self, year_n, sources_in, gamma_n, value, title, inames):
        """Plot sources over time."""
        stypes = list(sources_in.keys())
        if value == "nominal":
            yformat = r"\$k (nominal)"
            sources = sources_in
        else:
            yformat = r"\$k (" + str(year_n[0]) + r"\$)"
            sources = {k: v / gamma_n[:-1] for k, v in sources_in.items()}
        fig, ax = self._stack_plot(year_n, inames, title, range(len(inames)), sources, stypes, "upper left", yformat)

        return fig

    def plot_taxes(self, year_n, T_n, M_n, gamma_n, value, title, inames):
        """Plot taxes over time."""
        style = {"income tax": "-", "Medicare": "-."}
        if value == "nominal":
            series = {"income tax": T_n, "Medicare": M_n}
            yformat = r"\$k (nominal)"
        else:
            series = {"income tax": T_n / gamma_n[:-1], "Medicare": M_n / gamma_n[:-1]}
            yformat = r"\$k (" + str(year_n[0]) + r"\$)"
        fig, ax = self._line_income_plot(year_n, series, style, title, yformat)

        return fig



import streamlit as st

import sskeys as kz
import owlbridge as owb


ret = kz.titleBar(":material/finance: Monte Carlo")

if ret is None or kz.caseHasNoPlan():
    st.info("Case(s) must be first created before running this page.")
else:
    if kz.getCaseKey("rateType") != "varying" or (
        kz.getCaseKey("varyingType") is None or "stochastic" not in kz.getCaseKey("varyingType")
    ):
        st.info("Rates must be set to *stochastic* or *histochastic* to run Monte Carlo simulations.")
    else:
        st.write("Generate a histogram of results obtained from running mutliple scenarios with stochastic rates.")
        col1, col2, col3, col4 = st.columns(4, gap="large", vertical_alignment="bottom")
        with col1:
            kz.initKey("MC_cases", 100)
            kz.getIntNum("Number of random instances", "MC_cases", step=10, max_value=10000)
        with col4:
            st.button("Run Simulation", on_click=owb.runMC, disabled=kz.caseIsNotMCReady())

    st.divider()
    fig = kz.getCaseKey("monteCarloPlot")
    if fig:
        col1, col2 = st.columns(2, gap="medium")
        owb.renderPlot(fig, col1)
        col2.code(kz.getCaseKey("monteCarloSummary"), language=None)



"""
Factory for creating plot backends.

Copyright &copy; 2025 - Martin-D. Lacasse

Disclaimers: This code is for educatonal purposes only and does not constitute financial advice.

"""

from .base import PlotBackend
from .matplotlib_backend import MatplotlibBackend
from .plotly_backend import PlotlyBackend


class PlotFactory:
    """Factory for creating plot backends."""

    @staticmethod
    def createBackend(backend_type: str) -> PlotBackend:
        """
        Create a plot backend of the specified type.

        Args:
            backend_type: Type of backend to create ("matplotlib" or "plotly")

        Returns:
            A PlotBackend instance

        Raises:
            ValueError: If backend_type is not a valid option
        """
        if backend_type == "matplotlib":
            return MatplotlibBackend()
        elif backend_type == "plotly":
            return PlotlyBackend()
        else:
            raise ValueError(f"Unknown backend type: {backend_type}")



import streamlit as st
import pandas as pd
from io import StringIO, BytesIO
from functools import wraps
from datetime import datetime, date
import importlib
import sys

sys.path.insert(0, "./src")
sys.path.insert(0, "../src")

import owlplanner as owl                      # noqa: E402
from owlplanner.rates import FROM, TO         # noqa: E402

import sskeys as kz         # noqa: E402
import progress             # noqa: E402


def hasMOSEK():
    spec = importlib.util.find_spec("mosek")
    return spec is not None


def createPlan():
    name = kz.currentCaseName()
    inames = [kz.getCaseKey("iname0")]
    description = kz.getCaseKey("description")
    yobs = [kz.getCaseKey("yob0")]
    life = [kz.getCaseKey("life0")]
    if kz.getCaseKey("status") == "married":
        inames.append(kz.getCaseKey("iname1"))
        yobs.append(kz.getCaseKey("yob1"))
        life.append(kz.getCaseKey("life1"))

    strio = StringIO()
    kz.storeCaseKey("logs", strio)
    try:
        plan = owl.Plan(inames, yobs, life, name,
                        verbose=True, logstreams=[strio, strio])
        kz.setCaseKey("plan", plan)
    except Exception as e:
        st.error(f"Failed creation of plan '{name}': {e}")
        return

    plan.setDescription(description)

    val = kz.getGlobalKey("plotGlobalBackend")
    if val:
        plan.setPlotBackend(val)

    # Set default plot value from case settings.
    plot_val = kz.getCaseKey("plots")
    if plot_val:
        plan.setDefaultPlots(plot_val)

    # Force to pull key and set profile if key was defined.
    if kz.getCaseKey("spendingProfile"):
        setProfile(None)

    if kz.getCaseKey("duplicate"):
        _setContributions(plan, "copy")
    else:
        resetTimeLists()

    st.toast(f"Created new case *'{name}'*. You can now move to the next page.")


def _checkPlan(func):
    """
    Decorator to check if plan was created properly.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        plan = kz.getCaseKey("plan")
        if plan is None:
            st.error(f"Plan not yet created. Preventing to execute method {func.__name__}().")
            return None
        return func(plan, *args, **kwargs)

    return wrapper


def prepareRun(plan):
    ni = 2 if kz.getCaseKey("status") == "married" else 1

    startDate = kz.getCaseKey("startDate")
    bal = kz.getAccountBalances(ni)
    try:
        plan.setAccountBalances(taxable=bal[0], taxDeferred=bal[1], taxFree=bal[2], startDate=startDate)
    except Exception as e:
        st.error(f"Setting account balances failed: {e}")
        return

    amounts, ages, indexed = kz.getFixedIncome(ni, "p")
    try:
        plan.setPension(amounts, ages, indexed)
    except Exception as e:
        st.error(f"Failed setting pensions: {e}")
        return

    amounts, ages, indexed = kz.getFixedIncome(ni, "ss")
    try:
        plan.setSocialSecurity(amounts, ages)
    except Exception as e:
        st.error(f"Failed setting social security: {e}")
        return

    if ni == 2:
        benfrac = [kz.getCaseKey("benf0"), kz.getCaseKey("benf1"), kz.getCaseKey("benf2")]
        try:
            plan.setBeneficiaryFractions(benfrac)
        except Exception as e:
            st.error(f"Failed setting beneficiary fractions: {e}")
            return

        surplusFrac = kz.getCaseKey("surplusFraction")
        try:
            plan.setSpousalDepositFraction(surplusFrac)
        except Exception as e:
            st.error(f"Failed setting beneficiary fractions: {e}")
            return

    plan.setDescription(kz.getCaseKey("description"))
    plan.setHeirsTaxRate(kz.getCaseKey("heirsTx"))
    plan.setDividendRate(kz.getCaseKey("divRate"))
    plan.setExpirationYearOBBBA(kz.getCaseKey("yOBBBA"))

    _setInterpolationMethod(plan)
    _setAllocationRatios(plan)
    _setRates(plan)
    _setContributions(plan, "set")


def runAllCases():
    currentCase = kz.currentCaseName()
    for case in kz.onlyCaseNames():
        # Being here, current case must be fine.
        if case != currentCase:
            kz.switchToCaseName(case)
            runPlan()
    kz.switchToCaseName(currentCase)


@_checkPlan
def runPlan(plan):
    prepareRun(plan)

    objective, options = kz.getSolveParameters()
    try:
        plan.solve(objective, options=options)
    except Exception as e:
        st.error(f"Solution failed: {e}")
        kz.storeCaseKey("caseStatus", "exception")
        kz.storeCaseKey("summaryDf", None)
        return

    kz.storeCaseKey("caseStatus", plan.caseStatus)
    if plan.caseStatus == "solved":
        kz.storeCaseKey("summaryDf", plan.summaryDf())
        kz.storeCaseKey("casetoml", getCaseString().getvalue())
    else:
        kz.storeCaseKey("summaryDf", None)
        kz.storeCaseKey("casetoml", "")


@_checkPlan
def runHistorical(plan):
    plan1 = owl.clone(plan)
    prepareRun(plan1)

    hyfrm = kz.getCaseKey("hyfrm")
    hyto = kz.getCaseKey("hyto")

    objective, options = kz.getSolveParameters()
    try:
        mybar = progress.Progress(None)
        fig, summary = plan1.runHistoricalRange(objective, options, hyfrm, hyto, figure=True, progcall=mybar)
        kz.storeCaseKey("histoPlot", fig)
        kz.storeCaseKey("histoSummary", summary)
    except Exception as e:
        kz.storeCaseKey("histoPlot", None)
        kz.storeCaseKey("histoSummary", None)
        st.error(f"Historical solution failed: {e}")
        return


@_checkPlan
def runMC(plan):
    plan1 = owl.clone(plan)
    prepareRun(plan1)

    N = kz.getCaseKey("MC_cases")

    objective, options = kz.getSolveParameters()
    try:
        mybar = progress.Progress(None)
        fig, summary = plan1.runMC(objective, options, N, figure=True, progcall=mybar)
        kz.storeCaseKey("monteCarloPlot", fig)
        kz.storeCaseKey("monteCarloSummary", summary)
    except Exception as e:
        kz.storeCaseKey("monteCarloPlot", None)
        kz.storeCaseKey("monteCarloSummary", None)
        st.error(f"Monte Carlo solution failed: {e}")


@_checkPlan
def setRates(plan):
    _setRates(plan)


def _setRates(plan):
    yfrm = kz.getCaseKey("yfrm")
    yto = kz.getCaseKey("yto")

    if kz.getCaseKey("rateType") == "fixed":
        if kz.getCaseKey("fixedType") == "historical average":
            plan.setRates("historical average", yfrm, yto)
            # Set fxRates back to computed values.
            for j in range(4):
                kz.storeCaseKey("fxRate" + str(j), 100 * plan.tau_kn[j, -1])
        else:
            plan.setRates(
                "user",
                values=[
                    float(kz.getCaseKey("fxRate0")),
                    float(kz.getCaseKey("fxRate1")),
                    float(kz.getCaseKey("fxRate2")),
                    float(kz.getCaseKey("fxRate3")),
                ],
            )
    else:
        varyingType = kz.getCaseKey("varyingType")
        if varyingType.startswith("histo"):
            if varyingType == "historical":
                yfrm2 = min(yfrm, TO - plan.N_n + 1)
                kz.storeCaseKey("yfrm", yfrm2)
                if yfrm != yfrm2:
                    yfrm = yfrm2
                    st.warning(f"Using {yfrm} as starting year.")
                yto = min(TO, yfrm + plan.N_n - 1)
                kz.storeCaseKey("yto", yto)
            plan.setRates(varyingType, yfrm, yto)
            mean, stdev, corr, covar = owl.getRatesDistributions(yfrm, yto, plan.mylog)
            for j in range(4):
                kz.storeCaseKey("mean" + str(j), 100 * mean[j])
                kz.storeCaseKey("stdev" + str(j), 100 * stdev[j])
            q = 1
            for k1 in range(plan.N_k):
                for k2 in range(k1 + 1, plan.N_k):
                    kz.storeCaseKey("corr" + str(q), corr[k1, k2])
                    q += 1

        elif varyingType == "stochastic":
            means = []
            stdev = []
            corr = []
            for kk in range(plan.N_k):
                means.append(kz.getCaseKey("mean" + str(kk)))
                stdev.append(kz.getCaseKey("stdev" + str(kk)))
            for q in range(1, 7):
                corr.append(kz.getCaseKey("corr" + str(q)))
            plan.setRates(varyingType, values=means, stdev=stdev, corr=corr)
        else:
            raise RuntimeError("Logic error in setRates()")

    return True


@_checkPlan
def showAllocations(plan):
    figures = plan.showAllocations(figure=True)
    st.divider()
    st.markdown("#### :orange[Asset Allocation]")
    # n = 3 if kz.getCaseKey("allocType") == "account" else 2
    n = 2
    c = 0
    cols = st.columns(n, gap="medium")
    for fig in figures:
        renderPlot(fig, cols[c])
        c = (c + 1) % n


@_checkPlan
def showProfile(plan, col):
    fig = plan.showProfile(figure=True)
    if fig:
        col.write("#### :orange[Spending Profile]")
        renderPlot(fig, col)


@_checkPlan
def showRates(plan, col):
    fig = plan.showRates(figure=True)
    if fig:
        col.write("#### :orange[Selected Rates Over Time Horizon]")
        renderPlot(fig, col)


@_checkPlan
def showRatesCorrelations(plan, col):
    fig = plan.showRatesCorrelations(figure=True)
    if fig:
        col.write("#### :orange[Correlations Between Return Rates]")
        renderPlot(fig, col)


@_checkPlan
def showIncome(plan):
    fig = plan.showIncome(figure=True)
    if fig:
        renderPlot(fig)


@_checkPlan
def showSources(plan):
    fig = plan.showSources(figure=True)
    if fig:
        renderPlot(fig)


@_checkPlan
def setInterpolationMethod(plan):
    _setInterpolationMethod(plan)


def _setInterpolationMethod(plan):
    plan.setInterpolationMethod(kz.getCaseKey("interpMethod"), kz.getCaseKey("interpCenter"),
                                kz.getCaseKey("interpWidth"))


@_checkPlan
def setContributions(plan, reset=True):
    _setContributions(plan, "reset")


def _setContributions(plan, action):
    """
    Set from UI -> Plan.
    """
    if kz.getCaseKey("timeList0") is None:
        return

    dicDf = {kz.getCaseKey("iname0"): kz.getCaseKey("timeList0")}
    if kz.getCaseKey("status") == "married":
        dicDf[kz.getCaseKey("iname1")] = kz.getCaseKey("timeList1")

    try:
        plan.readContributions(dicDf)
    except Exception as e:
        st.error(f"Failed to parse Wages and Contributions: {e}")
        return False

    if action == "copy":
        # Possible reconditionned data due to delta in year span.
        kz.setCaseKey("timeList0", plan.timeLists[kz.getCaseKey("iname0")])
        if kz.getCaseKey("status") == "married":
            kz.setCaseKey("timeList1", plan.timeLists[kz.getCaseKey("iname1")])
    elif action == "reset":
        kz.setCaseKey("timeListsFileName", "edited values")
    elif action == "set":
        kz.storeCaseKey("timeListsFileName", "edited values")

    plan.timeListsFileName = "edited values"


@_checkPlan
def readContributions(plan, stFile, file=None):
    """
    Set from file -> Plan -> UI.
    """
    if stFile is None:
        return False

    try:
        plan.readContributions(stFile)
    except Exception as e:
        st.error(f"Failed to parse Wages and Contributions file 'stFile.name': {e}")
        return False

    if file:
        name = file
    elif hasattr(stFile, "name"):
        name = stFile.name
    else:
        name = "unknown"

    kz.setCaseKey("stTimeLists", name)
    kz.setCaseKey("timeListsFileName", name)
    kz.setCaseKey("timeList0", plan.timeLists[kz.getCaseKey("iname0")])
    kz.setCaseKey("_timeList0", plan.timeLists[kz.getCaseKey("iname0")])
    if kz.getCaseKey("status") == "married":
        kz.setCaseKey("timeList1", plan.timeLists[kz.getCaseKey("iname1")])
        kz.setCaseKey("_timeList1", plan.timeLists[kz.getCaseKey("iname1")])
    plan.timeListsFileName = name

    return True


@_checkPlan
def resetContributions(plan):
    return plan.zeroContributions()


def resetTimeLists():
    tlists = resetContributions()
    for i, iname in enumerate(tlists):
        kz.setCaseKey("timeList" + str(i), tlists[iname])


@_checkPlan
def setAllocationRatios(plan):
    _setAllocationRatios(plan)


def _setAllocationRatios(plan):
    if kz.getCaseKey("allocType") == "individual":
        try:
            generic = kz.getIndividualAllocationRatios()
            plan.setAllocationRatios("individual", generic=generic)
        except Exception as e:
            st.error(f"Setting asset allocation failed: {e}")
            return
    elif kz.getCaseKey("allocType") == "account":
        try:
            acc = kz.getAccountAllocationRatios()
            plan.setAllocationRatios("account", taxable=acc[0], taxDeferred=acc[1], taxFree=acc[2])
        except Exception as e:
            st.error(f"Setting asset allocation failed: {e}")
            return
    else:
        st.error(f"Internal error: Unknown account type {kz.getCaseKey('allocType')}.")


@_checkPlan
def plotSingleResults(plan):
    c = 0
    n = 2
    cols = st.columns(n, gap="medium")
    fig = plan.showRates(figure=True)
    if fig:
        cols[c].write("#### :orange[Annual Rates]")
        renderPlot(fig, cols[c])
        c = (c + 1) % n

    fig = plan.showNetSpending(figure=True)
    if fig:
        cols[c].write("#### :orange[Net Available Spending]")
        renderPlot(fig, cols[c])
        c = (c + 1) % n

    fig = plan.showGrossIncome(figure=True)
    if fig:
        cols[c].write("#### :orange[Taxable Ordinary Income]")
        renderPlot(fig, cols[c])
        c = (c + 1) % n

    # st.divider()
    # cols = st.columns(n, gap="medium")
    fig = plan.showSources(figure=True)
    if fig:
        cols[c].write("#### :orange[Raw Income Sources]")
        renderPlot(fig, cols[c])
        c = (c + 1) % n

    fig = plan.showAccounts(figure=True)
    if fig:
        cols[c].write("#### :orange[Savings Balance]")
        renderPlot(fig, cols[c])
        c = (c + 1) % n

    fig = plan.showTaxes(figure=True)
    if fig:
        cols[c].write("#### :orange[Federal Taxes and Medicare (+IRMAA)]")
        renderPlot(fig, cols[c])
        c = (c + 1) % n

    c = 0
    figs = plan.showAssetComposition(figure=True)
    if figs:
        # st.divider()
        st.write("#### :orange[Asset Composition]")
        col1, col2, _ = st.columns([0.6, 0.2, 0.2], gap="medium")
        for fig in figs:
            if fig:
                renderPlot(fig, col1)
            else:
                col1.write("#\n<div style='text-align: center'> This plot is empty </div>",
                           unsafe_allow_html=True)
            # c = (c + 1) % n


@_checkPlan
def setProfile(plan, key):
    if key is not None:
        kz.setpull(key)
    else:
        kz.flagModified()
    profile = kz.getCaseKey("spendingProfile")
    survivor = kz.getCaseKey("survivor")
    dip = kz.getCaseKey("smileDip")
    increase = kz.getCaseKey("smileIncrease")
    delay = kz.getCaseKey("smileDelay")
    plan.setSpendingProfile(profile, survivor, dip, increase, delay)


@_checkPlan
def setDefaultPlots(plan, key):
    val = kz.storepull(key)
    plan.setDefaultPlots(val)


def setGlobalPlotBackend(key):
    val = kz.getGlobalKey(key)
    # Apply to all existing cases.
    for casename in kz.onlyCaseNames():
        plan = kz.getKeyInCase("plan", casename)
        if plan:
            plan.setPlotBackend(val)


@_checkPlan
def showWorkbook(plan):
    wb = plan.saveWorkbook(saveToFile=False)
    if wb is None:
        return

    currencySheets = ["Income", "Cash Flow", "Sources", "Accounts"]
    for name in wb.sheetnames:
        if name == "Summary":
            continue

        dollars = False
        for word in currencySheets:
            if word in name:
                dollars = True
                break

        ws = wb[name]
        df = pd.DataFrame(ws.values)
        new_header = df.iloc[0]
        df = df[1:]
        df.columns = new_header
        if dollars:
            colfor = {}
            for col in df.columns:
                if col == "year":
                    colfor[col] = st.column_config.NumberColumn(None, format="%d", width="small")
                else:
                    # colfor[col] = st.column_config.NumberColumn(None, format="$ %,.0f")
                    colfor[col] = st.column_config.NumberColumn(None, format="accounting", step=1)
        else:
            colfor = {}
            for col in df.columns:
                if col == "year":
                    colfor[col] = st.column_config.NumberColumn(None, format="%d", width="small")
                else:
                    colfor[col] = st.column_config.NumberColumn(None, format="%.3f")

        st.write(f"#### :orange[{name}]")
        st.dataframe(df.astype(str), width="stretch", column_config=colfor, hide_index=True)

        if dollars:
            st.caption("Values are in nominal $, rounded to the nearest dollar.")
        else:
            st.caption("Values are fractional.")


@_checkPlan
def saveWorkbook(plan):
    wb = plan.saveWorkbook(saveToFile=False)
    buffer = BytesIO()
    if wb is None:
        return buffer
    try:
        wb.save(buffer)
    except Exception as e:
        raise Exception(f"Unanticipated exception: {e}.") from e

    return buffer


@_checkPlan
def saveContributions(plan):
    wb = plan.saveContributions()
    buffer = BytesIO()
    if wb is None:
        return buffer
    try:
        wb.save(buffer)
    except Exception as e:
        raise Exception(f"Unanticipated exception: {e}.") from e

    return buffer


@_checkPlan
def getCaseString(plan):
    stringBuffer = StringIO()
    if kz.getSolveParameters() is None:
        return ""
    plan.saveConfig(stringBuffer)

    return stringBuffer


@_checkPlan
def saveCaseFile(plan):
    stringBuffer = getCaseString()
    encoded_data = stringBuffer.getvalue().encode("utf-8")

    return BytesIO(encoded_data)


def createCaseFromFile(strio):
    logstrio = StringIO()
    try:
        plan = owl.readConfig(strio, logstreams=[logstrio], readContributions=False)
    except Exception as e:
        st.error(f"Failed to parse case file: {e}")
        return "", {}

    name, mydic = genDic(plan)
    mydic["logs"] = logstrio

    val = kz.getGlobalKey("plotGlobalBackend")
    if val:
        plan.setPlotBackend(val)

    return name, mydic


def genDic(plan):
    accName = ["txbl", "txDef", "txFree"]
    dic = {}
    dic["plan"] = plan
    dic["name"] = plan._name
    dic["description"] = plan._description
    dic["summaryDf"] = None
    dic["casetoml"] = ""
    dic["caseStatus"] = "new"
    dic["status"] = ["unknown", "single", "married"][plan.N_i]
    # Prepend year if not there.
    tdate = plan.startDate.replace("/", "-").split("-")
    if len(tdate) == 2:
        mystartDate = str(date.today().year) + "-" + plan.startDate
    elif len(tdate) == 3:
        mystartDate = str(date.today().year) + "-" + tdate[-2] + "-" + tdate[-1]
    else:
        raise ValueError(f"Wrong date format: {plan.startDate}")
    try:
        startDate = datetime.strptime(mystartDate, "%Y-%m-%d").date()
    except Exception as e:
        raise ValueError(f"Wrong date format {plan.startDate}: {e}") from e
    dic["startDate"] = startDate
    dic["interpMethod"] = plan.interpMethod
    dic["interpCenter"] = plan.interpCenter
    dic["interpWidth"] = plan.interpWidth
    dic["spendingProfile"] = plan.spendingProfile
    if plan.spendingProfile == "smile":
        dic["smileDip"] = plan.smileDip
        dic["smileIncrease"] = plan.smileIncrease
        dic["smileDelay"] = plan.smileDelay
    else:
        dic["smileDip"] = 15
        dic["smileIncrease"] = 12
        dic["smileDelay"] = 0
    dic["survivor"] = 100 * plan.chi
    dic["divRate"] = 100 * plan.mu
    dic["heirsTx"] = 100 * plan.nu
    dic["yOBBBA"] = plan.yOBBBA
    dic["surplusFraction"] = plan.eta
    dic["plots"] = plan.defaultPlots
    dic["allocType"] = plan.ARCoord
    dic["timeListsFileName"] = plan.timeListsFileName
    for j1 in range(plan.N_j):
        dic["benf" + str(j1)] = plan.phi_j[j1]

    for i in range(plan.N_i):
        dic["iname" + str(i)] = plan.inames[i]
        dic["yob" + str(i)] = plan.yobs[i]
        dic["life" + str(i)] = plan.expectancy[i]
        dic["ssAge" + str(i)] = plan.ssecAges[i]
        dic["ssAmt" + str(i)] = plan.ssecAmounts[i] / 1000
        dic["pAge" + str(i)] = plan.pensionAges[i]
        dic["pAmt" + str(i)] = plan.pensionAmounts[i] / 1000
        dic["pIdx" + str(i)] = plan.pensionIsIndexed[i]
        for j1 in range(plan.N_j):
            dic[accName[j1] + str(i)] = plan.beta_ij[i, j1] / 1000

        if plan.ARCoord == "individual":
            for k1 in range(plan.N_k):
                dic[f"j3_init%{k1}_{i}"] = int(plan.boundsAR["generic"][i][0][k1])
                dic[f"j3_fin%{k1}_{i}"] = int(plan.boundsAR["generic"][i][1][k1])
        elif plan.ARCoord == "account":
            longAccName = ["taxable", "tax-deferred", "tax-free"]
            for j2 in range(3):
                for k2 in range(plan.N_k):
                    dic[f"j{j2}_init%{k2}_{i}"] = int(plan.boundsAR[longAccName[j2]][i][0][k2])
                    dic[f"j{j2}_fin%{k2}_{i}"] = int(plan.boundsAR[longAccName[j2]][i][1][k2])
        else:
            st.error("Only 'individual' and 'account' asset allocations are currently supported")
            return None

    optionKeys = list(plan.solverOptions)
    optList = ["netSpending", "maxRothConversion", "noRothConversions",
               "startRothConversions", "withMedicare", "bequest", "solver",
               "spendingSlack", "oppCostX", "xorConstraints", "withSCLoop",]
    for key in optList:
        if key in optionKeys:
            dic[key] = plan.solverOptions[key]

    if "previousMAGIs" in optionKeys:
        dic["MAGI0"] = plan.solverOptions["previousMAGIs"][0]
        dic["MAGI1"] = plan.solverOptions["previousMAGIs"][1]

    if plan.objective == "maxSpending":
        dic["objective"] = "Net spending"
    else:
        dic["objective"] = "Bequest"

    if plan.rateMethod in ["default", "conservative", "optimistic", "historical average", "user"]:
        dic["rateType"] = "fixed"
        dic["fixedType"] = plan.rateMethod
    elif plan.rateMethod in ["histochastic", "historical", "stochastic"]:
        dic["rateType"] = "varying"
        dic["varyingType"] = plan.rateMethod

    # Initialize in both cases.
    for k1 in range(plan.N_k):
        dic["fxRate" + str(k1)] = 100 * plan.rateValues[k1]

    if plan.rateMethod in ["historical average", "histochastic", "historical"]:
        dic["yfrm"] = plan.rateFrm
        dic["yto"] = plan.rateTo
    else:
        dic["yfrm"] = FROM
        # Rates availability are trailing by 1 year.
        dic["yto"] = date.today().year - 1

    if plan.rateMethod in ["stochastic", "histochastic"]:
        qq = 1
        for k1 in range(plan.N_k):
            dic["mean" + str(k1)] = 100 * plan.rateValues[k1]
            dic["stdev" + str(k1)] = 100 * plan.rateStdev[k1]
            for k2 in range(k1 + 1, plan.N_k):
                dic["corr" + str(qq)] = plan.rateCorr[k1, k2]
                qq += 1

    return plan._name, dic


@_checkPlan
def backYearsMAGI(plan):
    thisyear = date.today().year
    backyears = [0, 0]
    goatyear = min(plan.yobs)
    if thisyear - goatyear >= 65:
        backyears[0] = thisyear - 2
        backyears[1] = thisyear - 1
    elif thisyear - goatyear >= 64:
        backyears[1] = thisyear - 1

    return backyears


def renderPlot(fig, col=None):
    """
    Render a plot using the appropriate Streamlit function based on the backend type

        Args:
        fig: The figure object from either matplotlib or plotly
        col: Optional Streamlit column to render in
    """
    if fig is None:
        return

    # Check if it's a plotly figure.
    if hasattr(fig, 'to_dict'):  # plotly figures have to_dict method.
        config = {"width": "stretch"}
        if col:
            col.plotly_chart(fig, config=config)
        else:
            st.plotly_chart(fig, config=config)
    else:  # matplotlib figure.
        if col:
            col.pyplot(fig)
        else:
            st.pyplot(fig)

    # Add a space below each figure
    if col:
        # col.write("####")
        col.divider()
    else:
        st.divider()
        # st.write("####")


def version():
    return owl.__version__
